from .database import recoBase
from pyqtgraph.Qt import QtGui, QtCore
from .connectedObjects import connectedBox, connectedCircle, boxCollection


class pixel2d(recoBase):

    """docstring for pixel2d"""

    def __init__(self):
        super(pixel2d, self).__init__()
        self._productName = 'pixel2d'
        self._product_id = 3

        self._listOfClusters = []

        # Defining the pixel2d colors:
        self._clusterColors = [
            (0, 147, 147, 125),  # dark teal
            (0, 0, 252, 125),   # bright blue
            (156, 0, 156, 125),  # purple
            (255, 0, 255, 125),  # pink
            (255, 0, 0, 125),  # red
            (175, 0, 0, 125),  # red/brown
            (252, 127, 0, 125),  # orange
            (102, 51, 0, 125),  # brown
            (127, 127, 127, 125),  # dark gray
            (210, 210, 210, 125),  # gray
            (100, 253, 0)  # bright green
        ]

    # this is the function that actually draws the cluster.
    def drawObjects(self, view_manager, io_manager, meta):

        #Get the list of pixel2d sets:
        event_pixel2d = io_manager.get_data(3, str(self._producerName))
        if self._producerName in io_manager.producer_list(1):
            hasROI = True
        else:
            hasROI = False

        if hasROI:    
            event_roi = io_manager.get_data(1, str(self._producerName))

        for view in view_manager.getViewPorts():
            colorIndex = 0
            # get the plane
            thisPlane = view.plane()

            # Get the pixel2d clusters for this plane:
            clusters = event_pixel2d.Pixel2DClusterArray(thisPlane)

            # extend the list of clusters
            self._listOfClusters.append([])

            for i in xrange(len(clusters)):
                cluster = clusters[i]
                # Now make the cluster
                cluster_box_coll = boxCollection()
                cluster_box_coll.setColor(self._clusterColors[colorIndex])
                cluster_box_coll.setPlane(thisPlane)

                # Keep track of the cluster for drawing management
                self._listOfClusters[thisPlane].append(cluster_box_coll)

                # Get the matching ROI information:
                _event_ID = cluster.ID()
                if hasROI:
                    label = event_roi.ROIArray().at(_event_ID).Type()
                    cluster_box_coll.setLabel(label)

                # draw the hits in this cluster:
                cluster_box_coll.drawHits(view, cluster)



                colorIndex += 1
                if colorIndex >= len(self._clusterColors):
                    colorIndex = 0

    def clearDrawnObjects(self, view_manager):
        i_plane = 0
        # erase the clusters
        for plane in self._listOfClusters:
            view = view_manager.getViewPorts()[i_plane]
            i_plane += 1
            for cluster in plane:
                cluster.clearHits(view)


        self._listOfClusters = []



from gui import gui
from pyqtgraph.Qt import QtGui, QtCore

from .recobox import recoBox


# Inherit the basic gui to extend it
# override the gui to give the lariat display special features:


class evdgui(gui):

    """special evd gui"""

    def __init__(self):
        super(evdgui, self).__init__()
        # self._event_manager.truthLabelChanged.connect(self.updateMessageBar)

    # override the initUI function to change things:
    def initUI(self):
        super(evdgui, self).initUI()
        # Change the name of the labels for lariat:
        self.update()

    # This function sets up the eastern widget
    def getEastLayout(self):
        # This function just makes a dummy eastern layout to use.
        label1 = QtGui.QLabel("LArCV")
        label2 = QtGui.QLabel("Display")
        font = label1.font()
        font.setBold(True)
        label1.setFont(font)
        label2.setFont(font)



        self._eastWidget = QtGui.QWidget()
        # This is the total layout
        self._eastLayout = QtGui.QVBoxLayout()
        # add the information sections:
        self._eastLayout.addWidget(label1)
        self._eastLayout.addWidget(label2)
        self._eastLayout.addStretch(1)



        # Now we get the list of items that are drawable:
        drawableProducts = self._event_manager.getDrawableProducts()
        # print drawableProducts
        self._listOfRecoBoxes = []
        for product in drawableProducts:
            thisBox = recoBox(self,
                              product,
                              drawableProducts[product][1],
                              self._event_manager.getProducers(
                                  drawableProducts[product][1]))
            self._listOfRecoBoxes.append(thisBox)
            thisBox.activated[str].connect(self.recoBoxHandler)
            self._eastLayout.addWidget(thisBox)
        self._eastLayout.addStretch(2)

        self._eastWidget.setLayout(self._eastLayout)
        self._eastWidget.setMaximumWidth(150)
        self._eastWidget.setMinimumWidth(100)
        return self._eastWidget

    def drawableProductsChanged(self):
        # self.removeItem(self._eastLayout)
        self._eastWidget.close()
        east = self.getEastLayout()
        self.slave.addWidget(east)
        self.update()

        # self._eastLayout.setVisible(False)
        # self._eastLayout.setVisible(True)


    def recoBoxHandler(self, text):
        sender = self.sender()
        # print sender.product(), "was changed to" , text
        if text == "--Select--" or text == "--None--":
            self._event_manager.redrawProduct(sender.name(), None, self._view_manager)
            return
        else:
            self._event_manager.redrawProduct(sender.name(), text, self._view_manager)



from .image2d    import image2d
from .pixel2d    import pixel2d
from .particle2d import particle2d
from .cluster2d  import cluster2d
from .sparse2d   import sparse2d
from .bbox2d     import bbox2d
# This is the class that maintains the list of drawable items.
# If your class isn't here, it can't be drawn
import collections


class drawableItems(object):

    """This class exists to enumerate the drawableItems"""
    # If you make a new drawing class, add it here

    def __init__(self):
        super(drawableItems, self).__init__()
        # items are stored as pointers to the classes (not instances)
        self._drawableClasses = collections.OrderedDict()
        self._drawableClasses.update({'Image2D':     [image2d, 'image2d']})
        self._drawableClasses.update({'Cluster2D':   [cluster2d, 'cluster2d']})
        self._drawableClasses.update({'Sparse2D':    [sparse2d, 'sparse2d']})
        # self._drawableClasses.update({'Particle 2D': [particle2d, 'particle']})
        self._drawableClasses.update({'BBox 2D':     [bbox2d, 'bbox2d']})

    def getListOfTitles(self):
        return self._drawableClasses.keys()

    def getListOfItems(self):
        return zip(*self._drawableClasses.values())[1]

    def getDict(self):
        return self._drawableClasses


from .sparse3d import sparse3d
from .cluster3d import cluster3d
from .particle3d import particle3d
class drawableItems3D(object):

    """This class exists to enumerate the drawableItems in 3D"""
    # If you make a new drawing class, add it here

    def __init__(self):
        super(drawableItems3D, self).__init__()
        # items are stored as pointers to the classes (not instances)
        self._drawableClasses = collections.OrderedDict()
        self._drawableClasses.update({'Voxel3d': [sparse3d,"sparse3d"]})
        self._drawableClasses.update({'Cluster3d': [cluster3d,"cluster3d"]})
        # self._drawableClasses.update({'Particle3D': [particle3d,"particle"]})

    def getListOfTitles(self):
        return self._drawableClasses.keys()

    def getListOfItems(self):
        return zip(*self._drawableClasses.values())[1]

    def getDict(self):
        return self._drawableClasses






from .database import recoBase3D
from pyqtgraph.Qt import QtGui, QtCore
import numpy
import pyqtgraph as pg
import pyqtgraph.opengl as gl

class particle3d(recoBase3D):

    """docstring for cluster"""

    def __init__(self):
        super(particle3d, self).__init__()
        self._product_name = 'particle'


        self._box_template = numpy.array([[ 0 , 0, 0],
                                          [ 1 , 0, 0],
                                          [ 1 , 1, 0],
                                          [ 0 , 1, 0],
                                          [ 0 , 0, 0],
                                          [ 0 , 0, 1],
                                          [ 1 , 0, 1],
                                          [ 1 , 1, 1],
                                          [ 0 , 1, 1],
                                          [ 0 , 0, 1],
                                          [ 1 , 0, 1],
                                          [ 1 , 0, 0],
                                          [ 1 , 1, 0],
                                          [ 1 , 1, 1],
                                          [ 0 , 1, 1],
                                          [ 0 , 1, 0]],
                                         dtype=float)



    # this is the function that actually draws the cluster.
    def drawObjects(self, view_manager, io_manager, meta):


        event_particle = io_manager.get_data(self._product_name, str(self._producerName))

        # # This section draws voxels onto the environment:
        for particle in event_particle.as_vector():
            box = particle.boundingbox_3d()

            # Can create a 3D box using line plots.


            pts = self._box_template.copy()

            pts[:,0] *= box.width()
            pts[:,1] *= box.height()
            pts[:,2] *= box.depth()

            pts[:,0] += box.center_x() - meta.min_x() - 0.5*box.width()
            pts[:,1] += box.center_y() - meta.min_y() - 0.5*box.height()
            pts[:,2] += box.center_z() - meta.min_z() - 0.5*box.depth()


            line = gl.GLLinePlotItem(pos=pts,color=(1.0,1.0,1.0,1.0), width=3)
            view_manager.getView().addItem(line)
            self._drawnObjects.append(line)
            # break

#             min_x
# min_y
# min_z
# max_x
# max_y
# max_z

        # self.redraw(view_manager)



    def redraw(self, view_manager):


        if self._gl_voxel_mesh is not None:
            view_manager.getView().removeItem(self._gl_voxel_mesh)
            self._gl_voxel_mesh = None

        verts, faces, colors = self.buildTriangleArray(self._id_summed_charge,
                                                       self._assigned_colors,
                                                       view_manager)


        #make a mesh item: 
        mesh = gl.GLMeshItem(vertexes=verts,
                             faces=faces,
                             faceColors=colors,
                             smooth=False)
        # mesh.setGLOptions("additive")        
        self._gl_voxel_mesh = mesh
        view_manager.getView().addItem(self._gl_voxel_mesh)

    #     self._drawnObjects = []
    #     for plane, view in view_manager.getViewPorts().iteritems():
    #         # get the plane
    #         # thisPlane = view.plane()
    #         self._drawnObjects.append([])

    #         for particle in event_particle.as_vector():

    #             # particle = event_particle.at(i)
    #             bounding_box = particle.boundingbox_2d(plane)

    #             left = meta.wire_to_col(bounding_box.min_x(), plane)
    #             right = meta.wire_to_col(bounding_box.max_x(), plane)
    #             top = meta.time_to_row(bounding_box.min_y(), plane)
    #             bottom = meta.time_to_row(bounding_box.max_y(), plane)

    #             r = QtGui.QGraphicsRectItem(left, bottom, (right-left), (top - bottom))
    #             r.setPen(pg.mkPen('r'))
    #             r.setBrush(pg.mkColor((0,0,0,0)))
    #             self._drawnObjects[plane].append(r)
    #             view._plot.addItem(r)

    #     return

    # # def clearDrawnObjects(self, view_manager):
    # #     i_plane = 0
    # #     # erase the clusters
    # #     for plane in self._listOfClusters:
    # #         view = view_manager.getViewPorts()[i_plane]
    # #         i_plane += 1
    # #         for cluster in plane:
    # #             cluster.clearHits(view)


    # #     self._listOfClusters = []



from .database import dataBase, recoBase
from .drawableItems import drawableItems
# from hit import hit
# from cluster import cluster
# from shower import shower
# from track import track
# from wire import wire, rawDigit, recoWire
# from match import match
# from endpoint2d import endpoint2d
# from vertex import vertex
# from mctruth import mctruth
# from spacepoint import spacepoint
import pyqtgraph.opengl as gl
from .drawableItems import drawableItems3D
#     from track import track3D
#     from shower import shower3D
#     from mctrack import mctrack3D
#     from spacepoint import spacepoint3D
#     from opflash import opflash3D





import sys, signal
import argparse
# import collections
from pyqtgraph.Qt import QtGui, QtCore
import pyqtgraph as pg
import numpy as np

# import evdmanager
from .view_manager import view_manager

class gui(QtGui.QWidget):

  def __init__(self):
    super(gui, self).__init__()

    # initUI should not do ANY data handling, it should only get the interface loaded
    self._view_manager = view_manager()

  def connect_manager(self,manager):
    self._event_manager = manager

  def closeEvent(self, event):
    self.quit()  

  def quit(self):
    # if self._running:
      # self.stopRun()
    QtCore.QCoreApplication.instance().quit()

  def metaChanged(self, meta):
    self._event_manager.refresh_meta()
    for plane, view in self._view_manager.getViewPorts().items():
      view.updateRange(meta)

  def update(self):
    # set the text boxes correctly:
    self._fileEntry.setText(str(self._event_manager.entry()))

    eventLabel = "Ev: " + str(self._event_manager.event())
    self._eventLabel.setText(eventLabel)
    runLabel = "Run: " + str(self._event_manager.run())
    self._runLabel.setText(runLabel)
    subrunLabel = "Subrun: " + str(self._event_manager.subrun())
    self._subrunLabel.setText(subrunLabel)
    self.metaChanged(self._event_manager.meta())
    
    self._event_manager.drawFresh(self._view_manager)

  # This function prepares the buttons such as prev, next, etc and returns a layout
  def getEventControlButtons(self):

    # This is a box to allow users to enter an event (larlite numbering)
    self._goToLabel = QtGui.QLabel("Go to: ")
    self._fileEntry = QtGui.QLineEdit()
    self._fileEntry.setToolTip("Enter an event to skip to that event (sequential numbering")
    self._fileEntry.returnPressed.connect(self.goToEventWorker)
    # These labels display current events
    self._runLabel = QtGui.QLabel("Run: 0")
    self._eventLabel = QtGui.QLabel("Ev.: 0")
    self._subrunLabel = QtGui.QLabel("Subrun: 0")

    # Jump to the next event
    self._nextButton = QtGui.QPushButton("Next")
    # self._nextButton.setStyleSheet("background-color: red")
    self._nextButton.clicked.connect(self._event_manager.next)
    self._nextButton.setToolTip("Move to the next event.")
    # Go to the previous event
    self._prevButton = QtGui.QPushButton("Previous")
    self._prevButton.clicked.connect(self._event_manager.prev)
    self._prevButton.setToolTip("Move to the previous event.")

    # pack the buttons into a box
    self._eventControlBox = QtGui.QVBoxLayout()

    # Make a horiztontal box for the event entry and label:
    self._eventGrid = QtGui.QHBoxLayout()
    self._eventGrid.addWidget(self._goToLabel)
    self._eventGrid.addWidget(self._fileEntry)
    # Another horizontal box for the run/subrun
    # self._runSubRunGrid = QtGui.QHBoxLayout()
    # self._runSubRunGrid.addWidget(self._eventLabel)
    # self._runSubRunGrid.addWidget(self._runLabel)
    # Pack it all together
    self._eventControlBox.addLayout(self._eventGrid)
    self._eventControlBox.addWidget(self._eventLabel)
    self._eventControlBox.addWidget(self._runLabel)
    self._eventControlBox.addWidget(self._subrunLabel)
    self._eventControlBox.addWidget(self._nextButton)
    self._eventControlBox.addWidget(self._prevButton)

    return self._eventControlBox
  

  # this function helps pass the entry of the line edit item to the event control
  def goToEventWorker(self):
    try:
      event = int(self._fileEntry.text())
    except:
      print("Error, must enter an integer")
      self._fileEntry.setText(str(self._event_manager.entry()))
      return
    self._event_manager.go_to_entry(event)

  # This function prepares the range controlling options and returns a layout
  def getDrawingControlButtons(self):

    # Button to set range to max
    self._maxRangeButton = QtGui.QPushButton("Max Range")
    self._maxRangeButton.setToolTip("Set the range of the viewers to show the whole event")
    self._maxRangeButton.clicked.connect(self._view_manager.setRangeToMax)

    self._lockAspectRatio = QtGui.QCheckBox("Lock A.R.")
    self._lockAspectRatio.setToolTip("Lock the aspect ratio to 1:1")
    self._lockAspectRatio.stateChanged.connect(self.lockARWorker)

    self._falseColorSelection = QtGui.QCheckBox("False Color")
    self._falseColorSelection.setToolTip("Applies a false color scheme to images")
    self._falseColorSelection.stateChanged.connect(self.falseColorWorker)

    # add a box to restore the drawing defaults:
    self._restoreDefaults = QtGui.QPushButton("Restore Defaults")
    self._restoreDefaults.setToolTip("Restore the drawing defaults of the views.")
    self._restoreDefaults.clicked.connect(self.restoreDefaultsWorker)

    # Pack the stuff into a layout

    self._drawingControlBox = QtGui.QVBoxLayout()
    self._drawingControlBox.addWidget(self._restoreDefaults)
    self._drawingControlBox.addWidget(self._maxRangeButton)
    self._drawingControlBox.addWidget(self._lockAspectRatio)
    self._drawingControlBox.addWidget(self._falseColorSelection)

    return self._drawingControlBox

  def viewSelectWorker(self):

    i = 0
    for i in range(self._event_manager.n_views()):
      if self.sender() == self._viewButtonArray[i]:
        self._view_manager.selectPlane(i)
        self._view_manager.refreshDrawListWidget()
        return
      else:
        i += 1

    # if there wasn't a match, then it must be the ALL button:
    if self.sender() != None:
      self._view_manager.selectPlane(-1)
      self._view_manager.refreshDrawListWidget()
      return


  def lockARWorker(self):
    if self._lockAspectRatio.isChecked():
      self._view_manager.lockAR(True)
    else:
      self._view_manager.lockAR(False)

  def falseColorWorker(self):
    if self._falseColorSelection.isChecked():
      for plane, view in self._view_manager.getViewPorts().items():
        view.falseColor(True)
    else:
      for plane, view in self._view_manager.getViewPorts().items():
        view.falseColor(False)

  def restoreDefaultsWorker(self):
    self._view_manager.restoreDefaults()
    self._view_manager.setRangeToMax()
    
  # This function prepares the quit buttons layout and returns it
  def getQuitLayout(self):
    self._quitButton = QtGui.QPushButton("Quit")
    self._quitButton.setToolTip("Close the viewer.")
    self._quitButton.clicked.connect(self.quit)
    return self._quitButton

  # This function combines the control button layouts, range layouts, and quit button
  def getWestLayout(self):

    event_control = self.getEventControlButtons()
    draw_control = self.getDrawingControlButtons()


    # Add the quit button?
    quit_control = self.getQuitLayout()
    
    self._westLayout = QtGui.QVBoxLayout()
    self._westLayout.addLayout(event_control)
    self._westLayout.addStretch(1)
    self._westLayout.addLayout(draw_control)
    self._westLayout.addStretch(1)

    # Add a section to allow users to just view one window instead of two/three
    self._viewButtonGroup = QtGui.QButtonGroup()
    # Draw all planes:
    self._allViewsButton = QtGui.QRadioButton("All")
    self._allViewsButton.clicked.connect(self.viewSelectWorker)
    self._viewButtonGroup.addButton(self._allViewsButton)

    # Put the buttons in a layout
    self._viewChoiceLayout = QtGui.QVBoxLayout()

    # Make a label for this stuff:
    self._viewChoiceLabel = QtGui.QLabel("View Options")
    self._viewChoiceLayout.addWidget(self._viewChoiceLabel)
    self._viewChoiceLayout.addWidget(self._allViewsButton)

    i = 0
    self._viewButtonArray = []
    for plane in range(self._event_manager.n_views()):
      button = QtGui.QRadioButton("Plane" + str(i))
      i += 1
      self._viewButtonGroup.addButton(button)
      button.clicked.connect(self.viewSelectWorker)
      self._viewButtonArray.append(button)
      self._viewChoiceLayout.addWidget(button)

    self._westLayout.addLayout(self._viewChoiceLayout)

    self._westLayout.addStretch(1)

    self._westLayout.addWidget(quit_control)
    self._westWidget = QtGui.QWidget()
    self._westWidget.setLayout(self._westLayout)
    self._westWidget.setMaximumWidth(150)
    self._westWidget.setMinimumWidth(100)
    return self._westWidget


  def getSouthLayout(self):
    # This layout contains the status bar, message bar, and the capture screen buttons

    # The screen capture button:
    self._screenCaptureButton = QtGui.QPushButton("Capture Screen")
    self._screenCaptureButton.setToolTip("Capture the entire screen to file")
    self._screenCaptureButton.clicked.connect(self.screenCapture)
    self._southWidget = QtGui.QWidget()
    self._southLayout = QtGui.QHBoxLayout()
    # Add a status bar
    self._statusBar = QtGui.QStatusBar()
    self._statusBar.showMessage("Test message")
    self._southLayout.addWidget(self._statusBar)
    self._messageBar = QtGui.QStatusBar()
    self._southLayout.addWidget(self._messageBar)
    # self._southLayout.addStretch(1)
    self._southLayout.addWidget(self._screenCaptureButton)
    self._southWidget.setLayout(self._southLayout)

    return self._southWidget

  def updateMessageBar(self,message):
    # print "Received a message: {msg}".format(msg=message)
    self._messageBar.showMessage(message)

  def getEastLayout(self):
    # This function just makes a dummy eastern layout to use.
    label = QtGui.QLabel("Dummy")
    self._eastWidget = QtGui.QWidget()
    self._eastLayout = QtGui.QVBoxLayout()
    self._eastLayout.addWidget(label)
    self._eastLayout.addStretch(1)
    self._eastWidget.setLayout(self._eastLayout)
    self._eastWidget.setMaximumWidth(200)
    self._eastWidget.setMinimumWidth(100)
    return self._eastWidget

  def refreshEastLayout(self):
    east = getEastLayout()
    self._eastLayout.setVisible(False)
    self._eastLayout.setVisible(True)

  def refreshCenterView(self):

    # for child in self.centerWidget.children():
    #   print type(child)
    #   if type(child) == QtGui.QVBoxLayout:
    #     layout = child

    # print layout.children()
    # print layout

    widget = self._view_manager.getDrawListWidget()
    # for child in widget.children():
    #   print child

    # print widget
    # print layout

    # print layout.children()

    # for i in reversed(range(self.centerWidget.layout.count())): 
        # layout.itemAt(i).widget().setParent(None)

    self.centerWidget.setVisible(False)   
    self.centerWidget.setVisible(True)   

  def initUI(self):


    # Get all of the widgets:
    self.eastWidget  = self.getEastLayout()
    self.westWidget  = self.getWestLayout()
    self.southLayout = self.getSouthLayout()

    # Area to hold data:
    nviews = self._event_manager.meta().n_views()
    # nviews = self._baseData._nviews
    for i in range(0, nviews):
      # These boxes hold the wire/time views:
      self._view_manager.addEvdDrawer(i)

    self._view_manager.connectStatusBar(self._statusBar)

    self.centerWidget = self._view_manager.getDrawListWidget()

    # Put the layout together


    self.master = QtGui.QVBoxLayout()
    self.slave = QtGui.QHBoxLayout()
    self.slave.addWidget(self.westWidget)
    self.slave.addWidget(self.centerWidget)
    self.slave.addWidget(self.eastWidget)
    self.master.addLayout(self.slave)
    self.master.addWidget(self.southLayout)

    self.setLayout(self.master)    


    self.setGeometry(0, 0, 2400, 1600)
    self.setWindowTitle('Event Display')    
    self.setFocus()
    self.show()
    # self._view_manager.setRangeToMax()

  def keyPressEvent(self,e):
    if e.key() == QtCore.Qt.Key_N:
      self._event_manager.next()
      return
    if e.key() == QtCore.Qt.Key_P:
      self._event_manager.prev()
      return
    if e.key() == QtCore.Qt.Key_C:
      # print "C was pressed"
      if e.modifiers() and QtCore.Qt.ControlModifier :
        self.quit()
        return

    # if e.key() == QtCore.Qt.Key_C:
  #     self._dataListsAndLabels['Clusters'].setFocus()
    # if e.key() == QtCore.Qt.Key_H:
  #     self._dataListsAndLabels['Hits'].setFocus()

    if e.key() == QtCore.Qt.Key_R:
      self.setRangeToMax()
      return

    super(gui, self).keyPressEvent(e)

  def screenCapture(self):
    print("Screen Capture!")
    dialog = QtGui.QFileDialog()
    r = self._event_manager.run()
    e = self._event_manager.event()
    s = self._event_manager.subrun()
    name = "evd_" + self._geometry.name() + "_R" + str(r)
    name = name + "_S" + str(s)
    name = name + "_E" + str(e) + ".png"
    f = dialog.getSaveFileName(self,"Save File",name,
        "PNG (*.png);;JPG (*.jpg);;All Files (*)")

    if pg.Qt.QT_LIB == pg.Qt.PYQT4:
      pixmapImage = QtGui.QPixmap.grabWidget(self)
      pixmapImage.save(f,"PNG")
    else:
      pixmapImage = super(gui, self).grab()
      pixmapImage.save(f[0],"PNG")





from pyqtgraph.Qt import QtGui, QtCore

# Import the class that manages the view windows
from .viewport import viewport


class view_manager(QtCore.QObject):
  """This class manages a collection of viewports"""


  def __init__(self):
    super(view_manager, self).__init__()
    self._nviews = 0
    self._drawerList = dict()
    self._cmapList = []

    self._selectedPlane = -1

    
  def resetEvdDrawerList(self, plane):
    self._drawerList = dict()


  def addEvdDrawer(self,plane):
    self._drawerList.update({ plane : viewport(plane)})
    self._nviews += 1
  
  def selectPlane(self,plane):
    self._selectedPlane = plane


  def restoreDefaults(self):
    for view in self._drawerList:
      view.restoreDefaults()


  def getDrawListWidget(self):

    self._widgetList = []

    # loop through the list and add the drawing windows and their scale
    self._widget = QtGui.QWidget()
    self._layout = QtGui.QVBoxLayout()
    self._layout.setSpacing(0)
    # self._layout.setMargin(0)
    self._layout.setContentsMargins(0,0,0,0)

    self._planeWidgets = []
    for plane, view in self._drawerList.items():
      widget,layout = view.getWidget()
      self._planeWidgets.append(widget)
      self._layout.addWidget(widget,0)

    self._widget.setLayout(self._layout)

    return self._widget

  def refreshDrawListWidget(self):

    # Draw all planes:
    if self._selectedPlane == -1:
      i = 0
      for widget in self._planeWidgets:
        widget.setVisible(True)
        i += 1

    else:
      i = 0
      for widget in self._planeWidgets:
        if i == self._selectedPlane:
          widget.setVisible(True)
        else:
          widget.setVisible(False)
        i += 1


  def connectStatusBar(self,statusBar):
    for plane, view in self._drawerList.items():
      view.connectStatusBar(statusBar)


  def setRangeToMax(self):
    for plane, view in self._drawerList.items():
      view.setRangeToMax()

  def autoRange(self,event_manager):
    for view in self._drawerList:
      xRange,yRange = event_manager.getAutoRange(view.plane())
      view.autoRange(xRange,yRange)

  def lockAR(self, lockRatio):
    for plane, view in self._drawerList.items():
      view.lockRatio(lockRatio)

 


  def getViewPorts(self):
    return self._drawerList



from .gui import gui
from .view_manager import view_manager
from .evdgui import evdgui
    
from .viewport import viewport
try:
    import pyqtgraph.opengl as gl
    from .gui3D import gui3D
    from .evdgui3D import evdgui3D
except:
    pass



from pyqtgraph.Qt import QtCore
from datatypes import drawableItems3D

import larcv 

from .evd_manager_base import evd_manager_base
from .event_meta import event_meta3D

try:
    import pyqtgraph.opengl as gl
except:
    print("Need opengl for the 3D viewer! Exiting ...")
    exit()

class evd_manager_3D(evd_manager_base):

    """This class handles file I/O and drawing for 3D viewer"""


    def __init__(self, config, _file=None):
        super(evd_manager_3D, self).__init__(config, _file)
        self._drawableItems = drawableItems3D()
                

    def init_manager(self, _file):

        # For the larcv manager, using the IOManager to get at the data
        self._driver =  larcv.ProcessDriver('ProcessDriver')
        self._driver.configure(self._config)

        # Meta keeps track of information about number of planes, visible
        # regions, etc.:
        self._meta = event_meta3D()


        # Drawn classes is a list of things getting drawn, as well.
        self._drawnClasses = dict()


        if _file != None:
            # flist=larcv.VectorOfString()
            if type(_file) is list:
            #     for f in _file: flist.push_back(f)
                self._driver.override_input_file(_file)
            else:
                flist = []
                flist.append(_file)
                self._driver.override_input_file(flist)

        self._driver.initialize()
        self.go_to_entry(0)

        self.refresh_meta()
        

    def refresh_meta(self, producer=None):
        # Read in any of the image2d products if none is specified.
        # Use it's meta info to build up the meta for the viewer


        product = "sparse3d"
        producers = self._driver.io().producer_list(product)
        if len(producers) == 0:
            product = "cluster3d"
            producers = self._driver.io().producer_list(product)
        
        if len(producers) == 0:
            raise Exception("No Meta avialable to define viewer boundaries")

        producer = producers[0]



        data = self._driver.io().get_data(product, producer)
        if product == "sparse3d":
            # data = larcv.EventSparseTensor3D.to_sparse_tensor(data)            
            meta = data.sparse_tensor(0).meta()
        if product == "cluster3d":
            # data = larcv.EventSparseCluster3D.to_sparse_cluster(data)
            meta = data.sparse_cluster(0).meta()

        self._meta.refresh(meta)




    # this function is meant for the first request to draw an object or
    # when the producer changes
    def redrawProduct(self, product, producer, view_manager):
        
        # print "Received request to redraw ", product, " by ",producer
        # First, determine if there is a drawing process for this product:  
        
        if producer is None:
            if product in self._drawnClasses:
                self._drawnClasses[product].clearDrawnObjects(view_manager)
                self._drawnClasses.pop(product)
            return
        if product in self._drawnClasses:
            self._drawnClasses[product].setProducer(producer)
            self._drawnClasses[product].clearDrawnObjects(view_manager)
            self._drawnClasses[product].drawObjects(view_manager, self._driver.io(), self.meta())
            return

        # Now, draw the new product
        if product in self._drawableItems.getListOfTitles():
            # drawable items contains a reference to the class, so
            # instantiate it
            drawingClass=self._drawableItems.getDict()[product][0]()

            drawingClass.setProducer(producer)
            self._drawnClasses.update({product: drawingClass})

            # Need to process the event
            drawingClass.drawObjects(view_manager, self._driver.io(), self.meta())


    def clearAll(self, view_manager):
        for recoProduct in self._drawnClasses:
            self._drawnClasses[recoProduct].clearDrawnObjects(view_manager)
        # self.clearTruth()

    def drawFresh(self, view_manager):

        self.clearAll(view_manager)
        # Draw objects in a specific order defined by drawableItems
        order = self._drawableItems.getListOfTitles()
        # self.drawTruth()
        for item in order:
            if item in self._drawnClasses:
                self._drawnClasses[item].drawObjects(view_manager, self._driver.io(), self.meta())


    def refreshColors(self, view_manager):
        order = self._drawableItems.getListOfTitles()
        for item in order:
            if item in self._drawnClasses:
                self._drawnClasses[item].refresh(view_manager)


from .database import recoBase
from pyqtgraph.Qt import QtGui, QtCore
import numpy
try:
    import pyqtgraph.opengl as gl
except:
    print("Error, must have open gl to use this viewer.")
    exit(-1)

import larcv

class sparse3d(recoBase):

    """docstring for sparse3d"""

    def __init__(self):
        super(sparse3d, self).__init__()
        self._product_name = 'sparse3d'
        self._gl_voxel_mesh = None
        self._id_summed_charge = dict()
        self._meta = None

        self._box_template = numpy.array([[ 0 , 0, 0],
                                          [ 1 , 0, 0],
                                          [ 1 , 1, 0],
                                          [ 0 , 1, 0],
                                          [ 0 , 0, 1],
                                          [ 1 , 0, 1],
                                          [ 1 , 1, 1],
                                          [ 0 , 1, 1]],
                                         dtype=float)


        self._faces_template = numpy.array([[0, 1, 2],
                                            [0, 2, 3],
                                            [0, 1, 4],
                                            [1, 5, 4],
                                            [1, 2, 5],
                                            [2, 5, 6],
                                            [2, 3, 6],
                                            [3, 6, 7],
                                            [0, 3, 7],
                                            [0, 4, 7],
                                            [4, 5, 7],
                                            [5, 6, 7]])


    # this is the function that actually draws the cluster.
    def drawObjects(self, view_manager, io_manager, meta):

        #Get the list of sparse3d sets:
        event_voxel3d = io_manager.get_data(self._product_name, str(self._producerName))
        # event_voxel3d = larcv.EventSparseTensor3D.to_sparse_tensor(event_voxel3d)


        voxels = event_voxel3d.as_vector()[0]
        self._meta = voxels.meta() 

        # view_manager.getView().updateMeta(self._meta)



        self._id_summed_charge = dict()
        # # This section draws voxels onto the environment:
        for voxel in voxels.as_vector():

            if voxel.id() >= self._meta.total_voxels():
                continue
            if voxel.id() in self._id_summed_charge:
                self._id_summed_charge[voxel.id()] += voxel.value()
            else:
                self._id_summed_charge.update({voxel.id() : voxel.value()})
        self.redraw(view_manager)



        # self.setColors(view_manager.getLookupTable(), view_manager.getLevels())
        # self.redraw(view_manager)

    def redraw(self, view_manager):


        if self._gl_voxel_mesh is not None:
            view_manager.getView().removeItem(self._gl_voxel_mesh)
            self._gl_voxel_mesh = None


        verts, faces, colors = self.buildTriangleArray(self._id_summed_charge,
                                                       view_manager)

        #make a mesh item: 
        mesh = gl.GLMeshItem(vertexes=verts,
                             faces=faces,
                             faceColors=colors,
                             smooth=False)

        # mesh.setGLOptions("additive")        
        mesh.setGLOptions("translucent")        
        self._gl_voxel_mesh = mesh
        view_manager.getView().addItem(self._gl_voxel_mesh)

    def buildTriangleArray(self, id_summed_charge, view_manager):


        n_voxels = len(id_summed_charge)
        # # # Allocate enough memory for all 3 numpy arrays upfront:

        verts = numpy.zeros((n_voxels*8,3))
        faces = numpy.zeros((n_voxels*12,3), dtype=numpy.int)
        colors = numpy.zeros((n_voxels*12,4))

        i = 0
        for voxel_id in id_summed_charge:
            # Don't draw this pixel if it's below the threshold:
            if id_summed_charge[voxel_id] < view_manager.getLevels()[0]:
                continue


            this_color = self.getColor(view_manager.getLookupTable(),
                                       view_manager.getLevels(),
                                       id_summed_charge[voxel_id])


            colors[12*i: 12*(i+1)] = this_color

            this_verts = self.makeBox(voxel_id, self._meta)
            faces[12*i:12*(i+1)] = self._faces_template + 8*i
            verts[8*i:8*(i+1)] = this_verts


            i += 1

        colors = colors[0:i*12]
        faces  = faces[0:i*12]
        verts  = verts[0:i*8]


        return verts, faces, colors

    def makeBox(self, voxel_id, meta):
        verts_box = numpy.copy(self._box_template)
        #Scale all the points of the box to the right voxel size:
        verts_box[:,0] *= meta.voxel_dimensions(0)
        verts_box[:,1] *= meta.voxel_dimensions(1)
        verts_box[:,2] *= meta.voxel_dimensions(2)


        #Shift the points to put the center of the cube at (0,0,0)
        verts_box[:,0] -= 0.5*meta.voxel_dimensions(0)
        verts_box[:,1] -= 0.5*meta.voxel_dimensions(1)
        verts_box[:,2] -= 0.5*meta.voxel_dimensions(2)
        
        #Move the points to the right coordinate in this space
        verts_box[:,0] += meta.position(voxel_id, 0) - meta.origin(0)
        verts_box[:,1] += meta.position(voxel_id, 1) - meta.origin(1)
        verts_box[:,2] += meta.position(voxel_id, 2) - meta.origin(2)


        # color_arr = numpy.ndarray((12, 4))
        # color_arr[:] = [1,1,1,1]

        return verts_box


    def getColor(self, _lookupTable, _levels, _voxel_value ):
        _min = _levels[0]
        _max = _levels[1]

        if _voxel_value >= _max:
            # print "Max " + str(_voxel_value)
            return _lookupTable[-1]
        elif _voxel_value < _min:
            # print "Min "  + str(_voxel_value)
            return (0,0,0,0)
        else:
            index = 255*(_voxel_value - _min) / (_max - _min)
            return _lookupTable[int(index)]


    def clearDrawnObjects(self, view_manager):
        if self._gl_voxel_mesh is not None:
            view_manager.getView().removeItem(self._gl_voxel_mesh)

        self._gl_voxel_mesh = None
        self._meta = None
        self._id_summed_charge = dict()

    def refresh(self, view_manager):
        self.redraw(view_manager)


# Import the class that manages the view windows
from .viewport3D import viewport3D
from pyqtgraph.Qt import QtCore, QtGui
import pyqtgraph as pg

colorMap = {'ticks': [(1, (151, 30, 22, 100)),
                      (0.791, (0, 181, 226, 100)),
                      (0.645, (76, 140, 43, 100)),
                      (0.47, (0, 206, 24, 100)),
                      (0.33333, (254, 209, 65, 100)),
                      (0, (255, 255, 255, 100))],
            'mode': 'rgb'}

class view_manager3D(QtCore.QObject):
    """This class manages a collection of viewports"""

    refreshColors = QtCore.pyqtSignal()

    def __init__(self):
        super(view_manager3D, self).__init__()
        self._view = viewport3D()


        # Define some color collections:

        self._cmap = pg.GradientWidget(orientation='right')
        self._cmap.restoreState(colorMap)
        self._cmap.sigGradientChanged.connect(self.gradientChangeFinished)
        # self._cmap.sigGradientChangeFinished.connect(self.gradientChangeFinished)
        self._cmap.resize(1, 1)

        self._lookupTable = self._cmap.getLookupTable(255, alpha=0.75)

        # These boxes control the levels.
        self._upperLevel = QtGui.QLineEdit()
        self._lowerLevel = QtGui.QLineEdit()

        self._upperLevel.returnPressed.connect(self.colorsChanged)
        self._lowerLevel.returnPressed.connect(self.colorsChanged)

        self._lowerLevel.setText(str(0.0))
        self._upperLevel.setText(str(2.0))

        # Fix the maximum width of the widgets:
        self._upperLevel.setMaximumWidth(35)
        self._cmap.setMaximumWidth(25)
        self._lowerLevel.setMaximumWidth(35)



        self._layout = QtGui.QHBoxLayout()
        self._layout.addWidget(self._view)

        colors = QtGui.QVBoxLayout()
        colors.addWidget(self._upperLevel)
        colors.addWidget(self._cmap)
        colors.addWidget(self._lowerLevel)

        self._layout.addLayout(colors)

    def gradientChangeFinished(self):
        self._lookupTable = self._cmap.getLookupTable(255, alpha=0.75)
        self.refreshColors.emit()


    def getLookupTable(self):
        return self._lookupTable*(1./255)


    def colorsChanged(self):
        self.refreshColors.emit() 



    def getLayout(self):
        return self._layout


    def setRangeToMax(self):
        pass

    def getView(self):
        return self._view

    def getLevels(self):
        _max = float(self._upperLevel.text())
        _min = float(self._lowerLevel.text())
        return (_min, _max)

    def setCameraPosition(self, pos=None, distance=None, elevation=None, azimuth=None):
        if pos is not None:
            self._view.setCameraPos(pos)
        else:
            self._view.setCameraPosition(distance=distance,elevation=elevation,azimuth=azimuth)

    def setCenter(self, center=None):
        if center is not None:
            self._view.setCenter(center)

    def pan(self, dx, dy, dz, relative=False):
        pass

    def update(self):
        self._view.update()

    def restoreDefaults(self):
        print("restoreDefaults called but not implemented")



#!/usr/bin/env python


# from gui import evdgui
import argparse
import sys
import signal
from pyqtgraph.Qt import QtGui, QtCore
import larcv

from manager import evd_manager_2D

import os

# This is to allow key commands to work when focus is on a box

from gui import evdgui


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    sys.stderr.write('\r')
    sys.exit()


def main():

    parser = argparse.ArgumentParser(description='Python based event display.')
    parser.add_argument("-c", "-C", "--config", help="Optional config file override.")
    parser.add_argument('file', nargs='*', help="Optional input file to use")

    args = parser.parse_args()

    app = QtGui.QApplication(sys.argv)

    if args.config is None:
      print("No config supplied, using default configuration file.")
      args.config = larcv.ProcessDriver.default_config()


    # If a file was passed, give it to the manager:

    manager = evd_manager_2D(args.config, args.file)


    thisgui = evdgui()
    thisgui.connect_manager(manager)
    # manager.goToEvent(0)
    thisgui.initUI()

    manager.eventChanged.connect(thisgui.update)
    manager.metaRefreshed.connect(thisgui.metaChanged)


    signal.signal(signal.SIGINT, sigintHandler)
    timer = QtCore.QTimer()
    timer.start(500)  # You may change this if you wish.
    timer.timeout.connect(lambda: None)  # Let the interpreter run each 500 ms.

    app.exec_()
    # sys.exit(app.exec_())


if __name__ == '__main__':
    main()



from .event_meta import event_meta
from .evd_manager_base import evd_manager_base
from .evd_manager_2D import evd_manager_2D
from .evd_manager_3D import evd_manager_3D
try:
    import pyqtgraph.opengl as gl
    # from .evdmanager import evd_manager_3D
except:
    pass



from .database import recoBase
from pyqtgraph.Qt import QtGui, QtCore


import pyqtgraph, numpy
import larcv

class cluster2d(recoBase):

    """docstring for cluster2d"""

    def __init__(self):
        super(cluster2d, self).__init__()
        self._product_name = 'cluster2d'

        # self._listOfClusters = []

        # Defining the cluster2d colors:
        self._clusterColors = [
            (0, 147, 147, 125),    # dark teal
            (0, 0, 252, 125),      # bright blue
            (156, 0, 156, 125),    # purple
            (255, 0, 255, 125),    # pink
            (255, 0, 0, 125),      # red
            (175, 0, 0, 125),      # red/brown
            (252, 127, 0, 125),    # orange
            (102, 51, 0, 125),     # brown
            (100, 253, 0, 125),    # bright green
            (255, 248, 202, 125),  # tan
            (255, 248, 202, 125),  # tan
            (255, 148, 241, 125),  # pink
        ]

    # this is the function that actually draws the cluster.
    def drawObjects(self, view_manager, io_manager, meta):

        #Get the list of cluster2d sets:
        event_sparse_cluster = io_manager.get_data(self._product_name, str(self._producerName))
        # event_sparse_cluster = larcv.EventSparseCluster2D.to_sparse_cluster(event_sparse_cluster)
        # if self._producerName in io_manager.producer_list(self._product_name):
        #     hasROI = True
        # else:
        #     hasROI = False

        # if hasROI:
        #     event_roi = io_manager.get_data(self._product_name, str(self._producerName))

        pen = pyqtgraph.mkPen((0, 0, 0), width=1)

        for plane, view in view_manager.getViewPorts().items():
            
            self._drawnObjects.append([])

            colorIndex = 0

            # Get the cluster2d clusters for this plane:
            # try:
            sparse_clusters = event_sparse_cluster.sparse_cluster(plane)
            # except:
                # continue
            meta = sparse_clusters.meta()

            # extend the list of clusters
            # self._listOfClusters.append([])

            clusters = sparse_clusters.as_vector()
            for i, cluster  in enumerate(clusters):
                if i == len(clusters) - 1: break


                indexes = cluster.indexes()
                values  = cluster.values()
                if len(values) == 0: continue

                # # Reject all out-of-bounds indexes:
                in_bounds = indexes < meta.total_voxels()
                # oob = indexes >= meta.total_voxels()
                # Merge the bounds:
                # in_bounds = numpy.logical_and(in_bounds, oob)
                indexes = indexes[in_bounds]
                values  = values[in_bounds]

                dims = [meta.number_of_voxels(0), meta.number_of_voxels(1) ]
                x, y = numpy.unravel_index(indexes, dims)

                y = y.astype(float) + 0.5
                x = x.astype(float) + 0.5

                # colors = cmap.map(values, mode='float')


                this_item = pyqtgraph.ScatterPlotItem()
                color = pyqtgraph.mkColor(self._clusterColors[colorIndex])

                this_item.setData(x, y, size=1, pxMode=False, 
                    symbol='s', pen=None, 
                    brush = color,
                    # hoverable=True, hoverPen=pen
                    )

                view._plot.addItem(this_item)
                self._drawnObjects[plane].append(this_item)




                colorIndex += 1
                if colorIndex >= len(self._clusterColors):
                    colorIndex = 0

    # def clearDrawnObjects(self, view_manager):
    #     i_plane = 0
    #     # erase the clusters
    #     for plane in self._listOfClusters:
    #         view = view_manager.getViewPorts()[i_plane]
    #         i_plane += 1
    #         for cluster in plane:
    #             cluster.clearHits(view)


    #     self._listOfClusters = []




class event_meta(object):

    def __init__(self):
        super(event_meta, self).__init__()

    def n_views(self):
        return max(self._n_views, 1)

    def refresh(self, meta_vec):


        self._n_views = len(meta_vec)

        self._x_min = []
        self._y_min = []
        self._x_max = []
        self._y_max = []
        self._y_n_pixels = []
        self._x_n_pixels = []
        x_ind = 0
        y_ind = 1
        for meta in meta_vec:
            self._x_min.append(meta.origin(0))
            self._y_min.append(meta.origin(1))
            self._x_max.append(meta.image_size(0) + meta.origin(0))
            self._y_max.append(meta.image_size(1) + meta.origin(1))
            self._x_n_pixels.append(meta.number_of_voxels(0))
            self._y_n_pixels.append(meta.number_of_voxels(1))

        for i in range(self._n_views):
            if self._x_min[i] == self._x_max[i]:
                self._x_max[i] = self._x_min[i] + self._x_n_pixels[i]
            if self._y_min[i] == self._y_max[i]:
                self._y_max[i] = self._y_min[i] + self._y_n_pixels[i]


    def cols(self, plane):
        return self._x_n_pixels[plane]

    def width(self, plane):
        return self._x_max[plane] - self._x_min[plane]

    def comp_x(self, plane):
        return self.width(plane) / self.cols(plane)

    def rows(self, plane):
        return self._y_n_pixels[plane]

    def height(self, plane):
        return self._y_max[plane] - self._y_min[plane]

    def comp_y(self, plane):
        return self.height(plane) / self.rows(plane)

    def wire_to_col(self, wire, plane):
        return self.cols(plane) * (1.0*(wire - self.min_x(plane)) / self.width(plane))

    def time_to_row(self, time, plane):
        return self.rows(plane) * (1.0*(time - self.min_y(plane)) / self.height(plane))


    def min_y(self, plane):
        return self._y_min[plane]

    def max_y(self, plane):
        return self._y_max[plane]

    def min_x(self, plane):
        return self._x_min[plane]

    def max_x(self, plane):
        return self._x_max[plane]



    def range(self, plane):
        if plane >= 0 and plane < self._n_views:
            return ((self._x_min[plane], self._x_min[plane] ),
                    (self._x_min[plane], self._x_min[plane]))
        else:
            print("ERROR: plane {} not available.".format(plane))
            return ((-1, 1), (-1, 1))

class event_meta3D(object):
    def __init__(self):
        super(event_meta3D, self).__init__()

    def refresh(self, meta):


        x_ind = 0
        y_ind = 1
        z_ind = 2
        self._x_min = meta.origin(x_ind)
        self._y_min = meta.origin(y_ind)
        self._z_min = meta.origin(z_ind)
        self._x_max = meta.image_size(x_ind) + meta.origin(x_ind)
        self._y_max = meta.image_size(y_ind) + meta.origin(y_ind)
        self._z_max = meta.image_size(z_ind) + meta.origin(z_ind)
        self._y_n_pixels = meta.number_of_voxels(x_ind)
        self._x_n_pixels = meta.number_of_voxels(y_ind)
        self._z_n_pixels = meta.number_of_voxels(z_ind)



    def size_voxel_x(self):
        return (self._x_max - self._x_min) / self._x_n_pixels

    def size_voxel_y(self):
        return (self._y_max - self._y_min) / self._y_n_pixels

    def size_voxel_z(self):
        return (self._z_max - self._z_min) / self._z_n_pixels

    def n_voxels_x(self):
        return self._x_n_pixels

    def n_voxels_y(self):
        return self._y_n_pixels

    def n_voxels_z(self):
        return self._z_n_pixels

    def dim_x(self):
        return self._x_max - self._x_min

    def dim_y(self):
        return self._y_max - self._y_min

    def dim_z(self):
        return self._z_max - self._z_min

    def width(self):
        return self.dim_x()

    def height(self):
        return self.dim_y()

    def length(self):
        return self.dim_z()

    def min_y(self):
        return self._y_min

    def max_y(self):
        return self._y_max

    def min_x(self):
        return self._x_min

    def max_x(self):
        return self._x_max

    def min_z(self):
        return self._z_min

    def max_z(self):
        return self._z_max





import sys, signal
import argparse
# import collections
from pyqtgraph.Qt import QtGui, QtCore
import pyqtgraph as pg
import numpy as np

from .view_manager3D import view_manager3D

# Wrap the spin box class to allow key signals to pass to the gui

class ConnectedSpinBox(QtGui.QSpinBox):
    """docstring for ConnectedSpinBox"""
    quitRequested = QtCore.pyqtSignal()
    def __init__(self):
        super(ConnectedSpinBox, self).__init__()
        

    def keyPressEvent(self, e):
        if e.key() == QtCore.Qt.Key_C:
            # print "C was pressed"
            if e.modifiers() and QtCore.Qt.ControlModifier :
                self.quitRequested.emit()
                return
        else:
            super(ConnectedSpinBox,self).keyPressEvent(e )



class gui3D(QtGui.QWidget):

  def __init__(self):
    super(gui3D, self).__init__()

    # initUI should not do ANY data handling, it should only get the interface loaded
    self._view_manager = view_manager3D()
    # self.setStyleSheet("background-color:rgb(230,230,230);")

  def connect_manager(self,manager):
    self._event_manager = manager

  def closeEvent(self, event):
    self.quit()  

  def quit(self):
    # if self._running:
      # self.stopRun()
    QtCore.QCoreApplication.instance().quit()


  def update(self):
    # set the text boxes correctly:
    self._entryBox.setText(str(self._event_manager.entry()))

    eventLabel = "Ev: " + str(self._event_manager.event())
    self._eventLabel.setText(eventLabel)
    runLabel = "Run: " + str(self._event_manager.run())
    self._runLabel.setText(runLabel)
    subrunLabel = "Subrun: " + str(self._event_manager.subrun())
    self._subrunLabel.setText(subrunLabel)
    self.metaChanged(self._event_manager.meta())
    self._event_manager.drawFresh(self._view_manager)
    

  def updateCameraInfo(self, cameraPos=None,worldPos=None):

    # This function is not for changing the view when the text boxes are changed
    # It is for keeping the text boxes in sync with the view as the user changes the view.

    # UPdate all of the camera text entries:
    if cameraPos is None:
        cameraPos = self._view_manager.getView().cameraPosition()
    if worldPos is None:
        worldPos = self._view_manager.getView().worldCenter()

    # To actually update these things corrently, we have to unplug 
    # the signals from the slots, update the fields, and then plug everything back in
    
    # print self._cameraCenterX.valueChangedf[]


    try:
        self._cameraCenterX.valueChanged.disconnect()
    except:
        pass
    try:
        self._cameraCenterY.valueChanged.disconnect()
    except:
        pass
    try:
        self._cameraCenterZ.valueChanged.disconnect()
    except:
        pass
    try:
        self._worldCenterX.valueChanged.disconnect()
    except:
        pass
    try:
        self._worldCenterY.valueChanged.disconnect()
    except:
        pass
    try:
        self._worldCenterZ.valueChanged.disconnect()
    except:
        pass

    self._cameraCenterX.setValue(cameraPos.x())
    self._cameraCenterY.setValue(cameraPos.y())
    self._cameraCenterZ.setValue(cameraPos.z())
    self._worldCenterX.setValue(worldPos.x())
    self._worldCenterY.setValue(worldPos.y())
    self._worldCenterZ.setValue(worldPos.z())

    self._cameraCenterX.valueChanged.connect(self.cameraCenterWorker)
    self._cameraCenterY.valueChanged.connect(self.cameraCenterWorker)
    self._cameraCenterZ.valueChanged.connect(self.cameraCenterWorker)
    self._worldCenterX.valueChanged.connect(self.worldCenterWorker)
    self._worldCenterY.valueChanged.connect(self.worldCenterWorker)
    self._worldCenterZ.valueChanged.connect(self.worldCenterWorker)


  # This function prepares the buttons such as prev, next, etc and returns a layout
  def getEventControlButtons(self):

    # This is a box to allow users to enter an event (larlite numbering)
    self._goToLabel = QtGui.QLabel("Go to: ")
    self._entryBox = QtGui.QLineEdit()
    self._entryBox.setToolTip("Enter an event to skip to that event (larlite numbering")
    self._entryBox.returnPressed.connect(self.goToEventWorker)
    # These labels display current events
    self._runLabel = QtGui.QLabel("Run: 0")
    self._eventLabel = QtGui.QLabel("Ev.: 0")
    self._subrunLabel = QtGui.QLabel("Subrun: 0")

    # Jump to the next event
    self._nextButton = QtGui.QPushButton("Next")
    # self._nextButton.setStyleSheet("background-color: red")
    self._nextButton.clicked.connect(self._event_manager.next)
    self._nextButton.setToolTip("Move to the next event.")
    # Go to the previous event
    self._prevButton = QtGui.QPushButton("Previous")
    self._prevButton.clicked.connect(self._event_manager.prev)
    self._prevButton.setToolTip("Move to the previous event.")

    
    # pack the buttons into a box
    self._eventControlBox = QtGui.QVBoxLayout()

    # Make a horiztontal box for the event entry and label:
    self._eventGrid = QtGui.QHBoxLayout()
    self._eventGrid.addWidget(self._goToLabel)
    self._eventGrid.addWidget(self._entryBox)
    # Another horizontal box for the run/subrun
    # self._runSubRunGrid = QtGui.QHBoxLayout()
    # self._runSubRunGrid.addWidget(self._eventLabel)
    # self._runSubRunGrid.addWidget(self._runLabel)
    # Pack it all together
    self._eventControlBox.addLayout(self._eventGrid)
    self._eventControlBox.addWidget(self._eventLabel)
    self._eventControlBox.addWidget(self._runLabel)
    self._eventControlBox.addWidget(self._subrunLabel)
    self._eventControlBox.addWidget(self._nextButton)
    self._eventControlBox.addWidget(self._prevButton)

    return self._eventControlBox
  

  # this function helps pass the entry of the line edit item to the event control
  def goToEventWorker(self):
    try:
      event = int(self._entryBox.text())
    except:
      print("Error, must enter an integer")
      self._entryBox.setText(str(self._event_manager.entry()))
      return
    self._event_manager.go_to_entry(event)

  # This function prepares the range controlling options and returns a layout
  def getDrawingControlButtons(self):

    # Button to set range to max
    self._maxRangeButton = QtGui.QPushButton("Max Range")
    self._maxRangeButton.setToolTip("Set the range of the viewers to show the whole event")
    self._maxRangeButton.clicked.connect(self._view_manager.setRangeToMax)


    # add a box to restore the drawing defaults:
    self._restoreDefaults = QtGui.QPushButton("Restore Defaults")
    self._restoreDefaults.setToolTip("Restore the drawing defaults of the views.")
    self._restoreDefaults.clicked.connect(self.restoreDefaultsWorker)


    # self._presetUButton = QtGui.QPushButton("U Plane View")
    # self._presetUButton.setToolTip("Set Camera to mimic the U Plane View")
    # self._presetUButton.clicked.connect(self.goToPresetCameraPosition)
    # self._presetVButton = QtGui.QPushButton("V Plane View")
    # self._presetVButton.setToolTip("Set Camera to mimic the V Plane View")
    # self._presetVButton.clicked.connect(self.goToPresetCameraPosition)
    # self._presetYButton = QtGui.QPushButton("Y Plane View")
    # self._presetYButton.setToolTip("Set Camera to mimic the Y Plane View")
    # self._presetYButton.clicked.connect(self.goToPresetCameraPosition)



    # Add some controls to manage the camera
    self._cameraControlLayout = QtGui.QHBoxLayout()

    # Get the min and max values for height, length, width:

    width  = self._event_manager.meta().max_x() - self._event_manager.meta().min_x()
    height = self._event_manager.meta().max_y() - self._event_manager.meta().min_y()
    length = self._event_manager.meta().max_z() - self._event_manager.meta().min_z()
    

    # Define the x,y,z location of the camera and world center
    self._cameraCenterLayout = QtGui.QVBoxLayout()
    self._cameraLabel = QtGui.QLabel("Camera")
    self._cameraCenterLayout.addWidget(self._cameraLabel)
    self._cameraCenterXLayout = QtGui.QHBoxLayout()
    self._cameraCenterXLabel = QtGui.QLabel("X:")
    self._cameraCenterX = ConnectedSpinBox()
    self._cameraCenterX.setValue(0)
    self._cameraCenterX.setRange(-10*width,10*width)
    self._cameraCenterX.quitRequested.connect(self.quit)
    self._cameraCenterX.valueChanged.connect(self.cameraCenterWorker)
    self._cameraCenterXLayout.addWidget(self._cameraCenterXLabel)
    self._cameraCenterXLayout.addWidget(self._cameraCenterX)

    self._cameraCenterLayout.addLayout(self._cameraCenterXLayout)
    self._cameraCenterYLayout = QtGui.QHBoxLayout()
    self._cameraCenterYLabel = QtGui.QLabel("Y:")
    self._cameraCenterY = ConnectedSpinBox()
    self._cameraCenterY.setValue(0)
    self._cameraCenterY.setRange(-10*height,10*height)
    self._cameraCenterY.quitRequested.connect(self.quit)
    self._cameraCenterY.valueChanged.connect(self.cameraCenterWorker)
    self._cameraCenterYLayout.addWidget(self._cameraCenterYLabel)
    self._cameraCenterYLayout.addWidget(self._cameraCenterY)

    self._cameraCenterLayout.addLayout(self._cameraCenterYLayout)
    self._cameraCenterZLayout = QtGui.QHBoxLayout()
    self._cameraCenterZLabel = QtGui.QLabel("Z:")
    self._cameraCenterZ = ConnectedSpinBox()
    self._cameraCenterZ.setValue(0)
    self._cameraCenterZ.setRange(-10*length,10*length)   
    self._cameraCenterZ.quitRequested.connect(self.quit)
    self._cameraCenterZ.valueChanged.connect(self.cameraCenterWorker)
    self._cameraCenterZLayout.addWidget(self._cameraCenterZLabel)
    self._cameraCenterZLayout.addWidget(self._cameraCenterZ)
    self._cameraCenterLayout.addLayout(self._cameraCenterZLayout)


    self._worldCenterLayout = QtGui.QVBoxLayout()
    self._worldLabel = QtGui.QLabel("world")
    self._worldCenterLayout.addWidget(self._worldLabel)
    self._worldCenterXLayout = QtGui.QHBoxLayout()
    self._worldCenterXLabel = QtGui.QLabel("X:")
    self._worldCenterX = ConnectedSpinBox()
    self._worldCenterX.setValue(0)
    self._worldCenterX.setRange(-10*width,10*width)
    self._worldCenterX.quitRequested.connect(self.quit)
    self._worldCenterX.valueChanged.connect(self.worldCenterWorker)
    self._worldCenterXLayout.addWidget(self._worldCenterXLabel)
    self._worldCenterXLayout.addWidget(self._worldCenterX)

    self._worldCenterLayout.addLayout(self._worldCenterXLayout)
    self._worldCenterYLayout = QtGui.QHBoxLayout()
    self._worldCenterYLabel = QtGui.QLabel("Y:")
    self._worldCenterY = ConnectedSpinBox()
    self._worldCenterY.setValue(0)
    self._worldCenterY.setRange(-10*height,10*height)
    self._worldCenterY.quitRequested.connect(self.quit)
    self._worldCenterY.valueChanged.connect(self.worldCenterWorker)
    self._worldCenterYLayout.addWidget(self._worldCenterYLabel)
    self._worldCenterYLayout.addWidget(self._worldCenterY)

    self._worldCenterLayout.addLayout(self._worldCenterYLayout)
    self._worldCenterZLayout = QtGui.QHBoxLayout()
    self._worldCenterZLabel = QtGui.QLabel("Z:")
    self._worldCenterZ = ConnectedSpinBox()
    self._worldCenterZ.setValue(0)
    self._worldCenterZ.setRange(-10*length,10*length)   
    self._worldCenterZ.quitRequested.connect(self.quit)
    self._worldCenterZ.valueChanged.connect(self.worldCenterWorker)
    self._worldCenterZLayout.addWidget(self._worldCenterZLabel)
    self._worldCenterZLayout.addWidget(self._worldCenterZ)
    self._worldCenterLayout.addLayout(self._worldCenterZLayout)



    # Pack the stuff into a layout

    self._drawingControlBox = QtGui.QVBoxLayout()
    self._drawingControlBox.addWidget(self._restoreDefaults)
    self._drawingControlBox.addWidget(self._maxRangeButton)
    self._drawingControlBox.addLayout(self._cameraControlLayout)

    self._drawingControlBox.addLayout(self._cameraCenterLayout)
    self._drawingControlBox.addLayout(self._worldCenterLayout)

    # self._drawingControlBox.addWidget(self._presetUButton)
    # self._drawingControlBox.addWidget(self._presetVButton)
    # self._drawingControlBox.addWidget(self._presetYButton)


    return self._drawingControlBox

  def worldCenterWorker(self):
    x = float(self._worldCenterX.text())
    y = float(self._worldCenterY.text())
    z = float(self._worldCenterZ.text())
    self._view_manager.setCenter((x,y,z))

  def cameraCenterWorker(self):
    # assemble the camera position:
    x = float(self._cameraCenterX.text())
    y = float(self._cameraCenterY.text())
    z = float(self._cameraCenterZ.text())
    # self.updateCameraInfo(cameraPos=QtGui.QVector3D(x,y,z))
    self._view_manager.setCameraPosition(pos = (x,y,z) )


  # def goToPresetCameraPosition(self):
  #   if self.sender() == self._presetUButton:
  #       self._view_manager.setCameraPosition(pos=(0, 500, -500))
  #       print "U Plane"
  #   if self.sender() == self._presetVButton:
  #       self._view_manager.setCameraPosition(pos=(0, 500, 500))
  #       print "V Plane"
  #   if self.sender() == self._presetYButton:
  #       self._view_manager.setCameraPosition(pos=(0, 500, 0))
  #       print "Y Plane"


  def restoreDefaultsWorker(self):
    self._view_manager.restoreDefaults()
    self._view_manager.setRangeToMax()
    
  # This function prepares the quit buttons layout and returns it
  def getQuitLayout(self):
    self._quitButton = QtGui.QPushButton("Quit")
    self._quitButton.setToolTip("Close the viewer.")
    self._quitButton.clicked.connect(self.quit)
    return self._quitButton

  # This function combines the control button layouts, range layouts, and quit button
  def getWestLayout(self):

    event_control = self.getEventControlButtons()
    draw_control = self.getDrawingControlButtons()


    # Add the quit button?
    quit_control = self.getQuitLayout()
    
    self._westLayout = QtGui.QVBoxLayout()
    self._westLayout.addLayout(event_control)
    self._westLayout.addStretch(1)
    self._westLayout.addLayout(draw_control)
    self._westLayout.addStretch(1)


    self._westLayout.addStretch(1)

    self._westLayout.addWidget(quit_control)
    self._westWidget = QtGui.QWidget()
    self._westWidget.setLayout(self._westLayout)
    self._westWidget.setMaximumWidth(150)
    self._westWidget.setMinimumWidth(100)
    return self._westWidget


  def getSouthLayout(self):
    # This layout contains the status bar and the capture screen buttons

    # The screen capture button:
    self._screenCaptureButton = QtGui.QPushButton("Capture Screen")
    self._screenCaptureButton.setToolTip("Capture the entire screen to file")
    self._screenCaptureButton.clicked.connect(self.screenCapture)
    self._southWidget = QtGui.QWidget()
    self._southLayout = QtGui.QHBoxLayout()
    # Add a status bar
    self._statusBar = QtGui.QStatusBar()
    self._statusBar.showMessage("Test message")
    self._southLayout.addWidget(self._statusBar)
    # self._southLayout.addStretch(1)
    self._southLayout.addWidget(self._screenCaptureButton)
    self._southWidget.setLayout(self._southLayout)

    return self._southWidget

  def getEastLayout(self):
    # This function just makes a dummy eastern layout to use.
    label = QtGui.QLabel("Dummy")
    self._eastWidget = QtGui.QWidget()
    self._eastLayout = QtGui.QVBoxLayout()
    self._eastLayout.addWidget(label)
    self._eastLayout.addStretch(1)
    self._eastWidget.setLayout(self._eastLayout)
    self._eastWidget.setMaximumWidth(200)
    self._eastWidget.setMinimumWidth(100)
    return self._eastWidget

  def refreshEastLayout(self):
    east = getEastLayout()
    self._eastLayout.setVisible(False)
    self._eastLayout.setVisible(True)

  def refreshCenterView(self):

    # for child in self.centerWidget.children():
    #   print type(child)
    #   if type(child) == QtGui.QVBoxLayout:
    #     layout = child

    # print layout.children()
    # print layout

    widget = self._view_manager.getDrawListWidget()
    # for child in widget.children():
    #   print child

    # print widget
    # print layout

    # print layout.children()

    # for i in reversed(range(self.centerWidget.layout.count())): 
        # layout.itemAt(i).widget().setParent(None)

    self.centerWidget.setVisible(False)   
    self.centerWidget.setVisible(True)   


  def metaChanged(self, meta):
    self._event_manager.refresh_meta()
    self._view_manager.getView().updateMeta(meta)


  def initUI(self):


    # Get all of the widgets:
    self.eastWidget  = self.getEastLayout()
    self.westWidget  = self.getWestLayout()
    self.southLayout = self.getSouthLayout()

    # Area to hold data:
    self._view = self._view_manager.getView()
    self._view.keyPressSignal.connect(self.keyPressEvent)
    self.centerLayout = self._view_manager.getLayout()
    self._view.quitRequested.connect(self.quit)
    self._view.viewChanged.connect(self.updateCameraInfo)

    # Put the layout together


    self.master = QtGui.QVBoxLayout()
    self.slave = QtGui.QHBoxLayout()
    self.slave.addWidget(self.westWidget)
    self.slave.addLayout(self.centerLayout)
    self.slave.addWidget(self.eastWidget)
    self.master.addLayout(self.slave)
    self.master.addWidget(self.southLayout)

    self.setLayout(self.master)    

    # ask the view manager to draw the planes:
    # self._view_manager.drawPlanes(self._event_manager)

    self.updateCameraInfo()

    self.setGeometry(0, 0, 2400, 1600)
    self.setWindowTitle('Event Display')    
    self.setFocus()
    self.show()
    self._view_manager.setRangeToMax()

  def keyPressEvent(self,e):
    if e.key() == QtCore.Qt.Key_N:
      self._event_manager.next()
      return
    if e.key() == QtCore.Qt.Key_P:
      self._event_manager.prev()
      return
    if e.key() == QtCore.Qt.Key_C:
      # print "C was pressed"
      if e.modifiers() and QtCore.Qt.ControlModifier :
        self.quit()
        return

    # if e.key() == QtCore.Qt.Key_C:
  #     self._dataListsAndLabels['Clusters'].setFocus()
    # if e.key() == QtCore.Qt.Key_H:
  #     self._dataListsAndLabels['Hits'].setFocus()

    if e.key() == QtCore.Qt.Key_R:
      self.setRangeToMax()
      return

    super(gui3D, self).keyPressEvent(e)

  def screenCapture(self):
    print("Screen Capture!")
    dialog = QtGui.QFileDialog()
    r = self._event_manager.run()
    e = self._event_manager.event()
    s = self._event_manager.subrun()
    name = "larcv_3D_" + "R" + str(r)
    name = name + "_S" + str(s)
    name = name + "_E" + str(e) + ".png"
    f = dialog.getSaveFileName(self,"Save File",name,
        "PNG (*.png);;JPG (*.jpg);;All Files (*)")

    # print filt
    # Print
    if (pg.Qt.QtVersion.startswith('4')):
      pixmapImage = QtGui.QPixmap.grabWidget(self)
      pixmapImage.save(f,"PNG")
    else:
      pixmapImage = super(gui3D, self).grab()
      pixmapImage.save(f[0],"PNG")






from .database import recoBase

import larcv
import copy
import numpy
class image2d(recoBase):

    """docstring for cluster"""

    def __init__(self):
        super(image2d, self).__init__()
        self._product_name = 'image2d'

    # this is the function that actually draws the cluster.
    def drawObjects(self, view_manager, io_manager, meta):


        image2d_array = io_manager.get_data(self._product_name, str(self._producerName))
        # image2d_array = larcv.EventImage2D.to_image2d(image2d_array)

 
        self._data_arr = []

        for i, image2d_plane in enumerate(image2d_array.as_vector()):
            try:
                thisView = view_manager.getViewPorts()[image2d_plane.meta().id()]
            except:
                thisView = view_manager.getViewPorts()[i]

            thisView.drawPlane(image2d_plane.as_array())

        return

    def clearDrawnObjects(self, view_manager):

        for index, view in view_manager.getViewPorts().items():
            view.drawBlank()



from pyqtgraph.Qt import QtGui, QtCore
import pyqtgraph as pg

# import line_profiler
# import atexit
# profile = line_profiler.LineProfiler()
# atexit.register(profile.print_stats)


# This class wraps the hit object to allow them to all function together
class connectedBox(QtGui.QGraphicsRectItem):

    """docstring for connectedBox"""

    def __init__(self, *args, **kwargs):
        super(connectedBox, self).__init__(*args)
        self.setAcceptHoverEvents(True)
        self._isHighlighted = False

    def hoverEnterEvent(self, e):
        self.setToolTip(self._ownerToolTip())
        self._ownerHoverEnter(e)

    def hoverLeaveEvent(self, e):
        self._ownerHoverExit(e)

    def mouseDoubleClickEvent(self, e):
        self._toggleHighlight()

    def connectOwnerHoverEnter(self, ownerHoverEnter):
        self._ownerHoverEnter = ownerHoverEnter

    def connectOwnerHoverExit(self, ownerHoverExit):
        self._ownerHoverExit = ownerHoverExit

    def connectToggleHighlight(self, ownerTH):
        self._toggleHighlight = ownerTH

    def connectToolTip(self, ownerToolTip):
        self._ownerToolTip = ownerToolTip


class connectedCircle(QtGui.QGraphicsEllipseItem):

    """docstring for connectedCircle"""

    def __init__(self, *args, **kwargs):
        super(connectedCircle, self).__init__(*args, **kwargs)
        self.setAcceptHoverEvents(True)
        self._isHighlighted = False

    def hoverEnterEvent(self, e):
        self.setToolTip(self._ownerToolTip())
        self._ownerHoverEnter(e)

    def hoverLeaveEvent(self, e):
        self._ownerHoverExit(e)

    def mouseDoubleClickEvent(self, e):
        self._toggleHighlight()

    def connectOwnerHoverEnter(self, ownerHoverEnter):
        self._ownerHoverEnter = ownerHoverEnter

    def connectOwnerHoverExit(self, ownerHoverExit):
        self._ownerHoverExit = ownerHoverExit

    def connectToggleHighlight(self, ownerTH):
        self._toggleHighlight = ownerTH

    def connectToolTip(self, ownerToolTip):
        self._ownerToolTip = ownerToolTip


class boxCollection(QtCore.QObject):
    # This class wraps a collection of hits and connects them together
    # it can draw and delete itself when provided with view_manage
    #
    mouseEnter = QtCore.pyqtSignal(QtGui.QGraphicsSceneHoverEvent)
    mouseExit = QtCore.pyqtSignal(QtGui.QGraphicsSceneHoverEvent)
    highlightChange = QtCore.pyqtSignal()

    def __init__(self):
        super(boxCollection, self).__init__()
        self._color = (0, 0, 0)
        self._plane = -1
        self._listOfHits = []
        self._isHighlighted = False
        self._acceptHoverEvents = False
        self._label = None

    def setColor(self, color):
        self._color = color

    def setPlane(self, plane):
        self._plane = plane

    def attachParams(self, params):
        self._params = params

    # Can connect boxCollections to other boxCollections or to cluster params
    def connect(self, other):
        self.mouseEnter.connect(other.hoverEnter)
        self.mouseExit.connect(other.hoverExit)
        self.highlightChange.connect(other.toggleHighlight)

    def setLabel(self, label):
        self._label = label

    def genToolTip(self):
        nhits = len(self._listOfHits)
        tip = "Hits: " + str(nhits)
        if self._label is not None:
            tip += "\nLabel: {}".format(self._label)
        return tip

    def hoverEnter(self, e):
        for hit in self._listOfHits:
            hit.setPen(pg.mkPen((0, 0, 0), width=1))
        # When the function is called from a box, the sender is none
        # When its passed from the params, the sender is something
        if self.sender() == None:
            self.mouseEnter.emit(e)

    def hoverExit(self, e):
        if self._isHighlighted:
            return
        for hit in self._listOfHits:
            hit.setPen(pg.mkPen(None))
        # When the function is called from a box, the sender is none
        # When its passed from the params, the sender is something
        if self.sender() == None:
            self.mouseExit.emit(e)

    def toggleHighlight(self):
        self._isHighlighted = not self._isHighlighted
        # When the function is called from a box, the sender is none
        # When its passed from the params, the sender is something
        if self.sender() == None:
            self.highlightChange.emit()

    # @profile
    def drawHits(self, view, cluster, meta):
        voxels = cluster.as_vector()
        color = pg.mkColor(self._color)
        pen = pg.mkPen(None)
        for voxel in voxels:
            # voxel = cluster.as_vector()[i]
            # Figure out the row and column from the id:
            col = meta.coordinate(voxel.id(),0)
            row = meta.coordinate(voxel.id(),1)
            # Draws a rectangle at (x,y,xlength, ylength)
            # print "Drawing voxel at ({}, {})".format(voxel.X(), voxel.Y())
            r = connectedBox(col, row, 1, 1) #, voxel.Width())
            r.setPen(pen)
            r.setBrush(color)
            self._listOfHits.append(r)
            view._plot.addItem(r)

            # Connect the voxel's actions with the clusters functions
            r.connectOwnerHoverEnter(self.hoverEnter)
            r.connectOwnerHoverExit(self.hoverExit)
            r.connectToggleHighlight(self.toggleHighlight)
            r.connectToolTip(self.genToolTip)

    def clearHits(self, view):
        for voxel in self._listOfHits:
            view._plot.removeItem(voxel)
        self._listOfHits = []
        self._label = None



from pyqtgraph.Qt import QtGui, QtCore
import pyqtgraph as pg

import numpy

class viewport(pg.GraphicsLayoutWidget):


  def __init__(self,plane=-1):
    super(viewport, self).__init__(border=None)

    # add a view box, which is a widget that allows an image to be shown
    self._plot = self.addPlot(border=None)

    # add an image item which handles drawing (and refreshing) the image
    self._item = pg.ImageItem(useOpenGL=True)

    self._plot.addItem(self._item)

    # # connect the scene to click events, used to get wires
    # self.scene().sigMouseClicked.connect(self.mouseClicked)
    # # connect the views to mouse move events, used to update the info box at the bottom
    self.scene().sigMouseMoved.connect(self.mouseMoved)
    self._plane = plane

    self._x_axis = self._plot.getAxis("bottom")
    self._y_axis = self._plot.getAxis("left")

    self._x_max_range = (-1, 1)
    self._y_max_range = (-1, 1)

    # Set up the blank data:
    self.setBackground('w')

    # each drawer contains its own color gradient and levels
    # this class can return a widget containing the right layout for everything
    # Define some color collections:

    self._bwMap = {'ticks': [(0, (0,0,0,255)),
                             (1, (255, 255, 255, 255))],
                   'mode' : 'rgb'}

    self._colorMap = {'ticks': [(0, (30, 30, 255, 255)),
                       (0.33333, (0, 255, 255, 255)),
                       (0.66666, (255,255,100,255)),
                       (1, (255, 0, 0, 255))],
                     'mode': 'rgb'}

    self._activeMap = self._bwMap

    self._cmap = pg.GradientWidget(orientation='right')
    self._cmap.restoreState(self._activeMap)
    self._cmap.sigGradientChanged.connect(self.refreshGradient)
    self._cmap.resize(1,1)

    # These boxes control the levels.
    self._upperLevel = QtGui.QLineEdit()
    self._lowerLevel = QtGui.QLineEdit()

    self._upperLevel.returnPressed.connect(self.levelChanged)
    self._lowerLevel.returnPressed.connect(self.levelChanged)

    self._lowerLevel.setText(str(-10))
    self._upperLevel.setText(str(10))


    # Fix the maximum width of the widgets:
    self._upperLevel.setMaximumWidth(35)
    self._cmap.setMaximumWidth(25)
    self._lowerLevel.setMaximumWidth(35)

    colors = QtGui.QVBoxLayout()
    colors.addWidget(self._upperLevel)
    colors.addWidget(self._cmap)
    colors.addWidget(self._lowerLevel)
    self._totalLayout = QtGui.QHBoxLayout()
    self._totalLayout.addWidget(self)
    self._totalLayout.addLayout(colors)
    self._widget = QtGui.QWidget()
    self._widget.setLayout(self._totalLayout)

  def falseColor(self, apply_fc):
    if apply_fc:
      self._activeMap = self._colorMap
    else:
      self._activeMap = self._bwMap
    self._cmap.restoreState(self._activeMap)
    self.refreshGradient()

  def restoreDefaults(self):
    self._lowerLevel.setText(str(self._geometry.getLevels(self._plane)[0]))
    self._upperLevel.setText(str(self._geometry.getLevels(self._plane)[1]))

    self._cmap.restoreState(self._activeMap)

  def mouseDrag(self):
    print( "mouse was dragged")

  def getWidget(self):
    return self._widget, self._totalLayout

  def levelChanged(self):
    # First, get the current values of the levels:
    lowerLevel = float(self._lowerLevel.text())
    upperLevel = float(self._upperLevel.text())

    # set the levels as requested:
    levels = (lowerLevel, upperLevel)
    # last, update the levels in the image:
    self._item.setLevels(levels)

  def refreshGradient(self):
    self._item.setLookupTable(self._cmap.getLookupTable(255))

  def mouseMoved(self, pos):


    self.q = self._item.mapFromScene(pos)
    self._lastPos = self.q
    if (pg.Qt.QtVersion.startswith('4')):
      message= QtCore.QString()
    else:
      message= str()
    if type(message) != str:
      message.append("X: ")
      message.append("{0:.1f}".format(self.q.x()))
    else:
      message += "X: "
      message += "{0:.1f}".format(self.q.x())
    if type(message) != str:
      message.append(", Y: ")
      message.append("{0:.1f}".format(self.q.y()))
    else:
      message += ", Y: "
      message += "{0:.1f}".format(self.q.y())


    # print message
    # if self.q.x() > 0 and self.q.x() < self._geometry.wRange(self._plane):
    #   if self.q.y() > 0 and self.q.y() < self._geometry.tRange():
    self._statusBar.showMessage(message)


  def generate_ticks_and_labels(self, n, minimum, maximum, step = 1):

    tick_vals = numpy.arange(n, step=step)

    tick_labels = numpy.arange(n, step=step, dtype=numpy.float32)


    tick_labels *= (maximum - minimum) / n
    tick_labels -= minimum

    return tick_vals, numpy.around(tick_labels)

  def updateRange(self, meta):

    # create the sets of ticks to update the image with:
    # Format is
    # [
    # [ (majorTickValue1, majorTickString1), (majorTickValue2, majorTickString2), ... ],
    # [ (minorTickValue1, minorTickString1), (minorTickValue2, minorTickString2), ... ],
    # ...
    # ]

    # Major ticks are the pixel widths, or 1
    # Minor ticks are 1
    # x_major_tick_vals   = numpy.arange(meta.cols(self._plane), step=1)


    # x_major_tick_labels = numpy.around(
    #   numpy.arange(meta.min_x(self._plane),
    #                meta.max_x(self._plane),
    #                meta.comp_x(self._plane)))

    x_major_tick_vals, x_major_tick_labels = self.generate_ticks_and_labels(
        n       = meta.cols(self._plane),
        minimum = meta.min_x(self._plane),
        maximum = meta.max_x(self._plane),
        step    = 1
      )


    self._x_max_range = (x_major_tick_vals[-1], x_major_tick_vals[0])


    x_major = numpy.column_stack((x_major_tick_vals, x_major_tick_labels))

    x_minor_tick_vals, x_minor_tick_labels = self.generate_ticks_and_labels(
        n       = meta.cols(self._plane),
        minimum = meta.min_x(self._plane),
        maximum = meta.max_x(self._plane),
        step    = 1./meta.comp_x(self._plane)
      )

    # x_minor_tick_vals = numpy.arange(meta.cols(self._plane), step=1./meta.comp_x(self._plane))
    # x_minor_tick_labels = numpy.arange(meta.min_x(self._plane), meta.max_x(self._plane), 1)

    # if len(x_minor_tick_vals) != len(x_minor_tick_labels):
    #   target = min(len(x_minor_tick_labels), len(x_minor_tick_vals))
    #   x_minor_tick_labels = x_minor_tick_labels[0:target]
    #   x_minor_tick_vals = x_minor_tick_vals[0:target]


    x_minor = numpy.column_stack((x_minor_tick_vals, x_minor_tick_labels))

    #Downsample the major ticks until they are 100x more than minor ticks:
    while x_minor.shape[0] / x_major.shape[0] < 100:
      x_major = x_major[::2,:]
      if len(x_major) < 10:
        break

    x_ticks = [x_major, x_minor]


    y_major_tick_vals, y_major_tick_labels = self.generate_ticks_and_labels(
        n       = meta.rows(self._plane),
        minimum = meta.min_y(self._plane),
        maximum = meta.max_y(self._plane),
        step    = 1
      )

    y_minor_tick_vals, y_minor_tick_labels = self.generate_ticks_and_labels(
        n       = meta.rows(self._plane),
        minimum = meta.min_y(self._plane),
        maximum = meta.max_y(self._plane),
        step    = 1./meta.comp_y(self._plane)
      )

    self._y_max_range = (y_major_tick_vals[-1], y_major_tick_vals[0])


    # y_major_tick_vals   = numpy.arange(meta.rows(self._plane), step=1)

    # y_tick_length = (meta.max_y(self._plane) - meta.max_y(self._plane)) / len(y_major_tick_vals)
    # y_major_tick_labels = numpy.copy(y_major_tick_vals)
    # y_major_tick_labels *= y_length

    # y_major_tick_labels = numpy.around(
    #   numpy.arange(meta.min_y(self._plane),
    #                meta.max_y(self._plane),
    #                meta.comp_y(self._plane)))


    y_major = numpy.column_stack((y_major_tick_vals, y_major_tick_labels))

    # y_minor_tick_vals = numpy.arange(meta.rows(self._plane), step=1./meta.comp_y(self._plane))
    # y_minor_tick_labels = numpy.arange(meta.min_y(self._plane),
    #                                    meta.max_y(self._plane), 1)

    # if y_minor_tick_labels.shape != y_minor_tick_vals.shape:
    #   y_minor_tick_vals = y_minor_tick_vals[0:len(y_minor_tick_labels)]
    
    y_minor = numpy.column_stack((y_minor_tick_vals, y_minor_tick_labels))

    #Downsample the major ticks until they are 100x more than minor ticks:
    while y_minor.shape[0] / y_major.shape[0] < 100:
      y_major = y_major[::2,:]
      if len(y_major) < 10:
        break


    y_ticks = [y_major, y_minor]

    self._x_axis.setTicks(x_ticks)
    self._y_axis.setTicks(y_ticks)

    self.setRangeToMax()

  def connectStatusBar(self, _statusBar):
    self._statusBar = _statusBar

  def setRangeToMax(self):
    self._plot.setRange(xRange=self._x_max_range,yRange=self._y_max_range, padding=0.002)



  def plane(self):
    return self._plane

  def lockRatio(self, lockAR ):
    ratio = (self._x_max_range[1] - self._x_max_range[0]) / (self._y_max_range[1] - self._y_max_range[0])
    if lockAR:
      self._plot.setAspectLocked(True, ratio=ratio)
    else:
      self._plot.setAspectLocked(False)

  def drawPlane(self, image):
    self._item.setImage(image,autoLevels=False)
    self._item.setLookupTable(self._cmap.getLookupTable(255))
    self._cmap.setVisible(True)
    self._upperLevel.setVisible(True)
    self._lowerLevel.setVisible(True)
    self._item.setVisible(False)
    self._item.setVisible(True)
    # Make sure the levels are actually set:
    self.levelChanged()

  def drawBlank(self):
    self._item.clear()
    self._cmap.setVisible(True)
    self._upperLevel.setVisible(True)
    self._lowerLevel.setVisible(True)








# from ROOT import evd


# These classes provide the basic interfaces for drawing objects
# It's meant to be inherited from for specific instances
# There is a class for drawing the following things:
#   - Raw wire data
#   - hits
#   - clusters
#   - nothing else ... yet

# To add a class, what you have to do is write a module to process it (larlite ana_processor)
# Put your class in RecoViewer or wherever.  I use the namespace evd for all of these classes
# so if you use something else, be sure to import it above from ROOT

# Your class needs to include getter methods for the data you need, and a setProducer function
# which sets the producer of the data type.  Then, write a python class that inherits from
# recoBase below.  Your class must override the function drawObjects to display it's data
# on the views.  Your class is responsible for maintaining the items being drawn, and must
# clear them with the clearDrawnObjects function.  Store drawn objects in
# _drawnObjects.

# Lastly, extend the drawableItems class with one line:
# self._drawableClasses.update({'dataProduct':yourPythonClassName})
# This gives the viewer the connection between the class you made and the
# product you want to draw

# Beyond that, inside your larlite or python class you can do pretty much whatever you want
# Run merging, matching, ertool, who cares.



class dataBase(object):

    """basic class from which data objects inherit"""

    def __init__(self):
        super(dataBase, self).__init__()
        self._productName = "null"
        self._producerName = "null"

    def productName(self):
        return self._productName

    def producerName(self):
        return self._producerName

    def setProducer(self, name):
        self._producerName = name


# Reco base, all reco objects must inherit from this
class recoBase(dataBase):

    """docstring for recoBase"""

    def __init__(self):
        super(recoBase, self).__init__()
        self._drawnObjects = []
        self._process = None

    def clearDrawnObjects(self, view_manager):
        for plane, view in view_manager.getViewPorts().items():
            if len(self._drawnObjects) > plane:
                for item in self._drawnObjects[plane]:
                    view._plot.removeItem(item)

        # clear the list:
        self._drawnObjects = []

    def getDrawnObjects(self):
        return self._drawnObjects

    def drawObjects(self, view_manager):
        pass

class recoBase3D(dataBase):

    """docstring for recoBase3D"""

    def __init__(self):
        super(recoBase3D, self).__init__()
        self._drawnObjects = []


    def clearDrawnObjects(self, view_manager):
        view = view_manager.getView()
        for item in self._drawnObjects:
            view.removeItem(item)
        # clear the list:
        self._drawnObjects = []

    def getDrawnObjects(self):
        return self._drawnObjects

    def drawObjects(self, view_manager):
        pass

    def refresh(self, view_manager):
        pass


from pyqtgraph.Qt import QtGui, QtCore


class ComboBoxWithKeyConnect(QtGui.QComboBox):

    def __init__(self):
        super(ComboBoxWithKeyConnect, self).__init__()

    def connectOwnerKPE(self, kpe):
        self._owner_KPE = kpe

    def keyPressEvent(self, e):
        if e.key() == QtCore.Qt.Key_Up:
            super(ComboBoxWithKeyConnect, self).keyPressEvent(e)
            return
        if e.key() == QtCore.Qt.Key_Down:
            super(ComboBoxWithKeyConnect, self).keyPressEvent(e)
            return
        else:
            self._owner_KPE(e)
        # if e.key() == QtCore.Qt.Key_N:
        #     self._owner_KPE(e)
        #     pass
        # if e.key() == QtCore.Qt.Key_P:
        #     self._owner_KPE(e)
        #     pass
        # else:
        #     super(ComboBoxWithKeyConnect, self).keyPressEvent(e)
        #     self._owner_KPE(e)

# This is a widget class that contains the label and combo box
# It also knows what to do when updating



class recoBox(QtGui.QWidget):
    activated = QtCore.pyqtSignal(str)

    """docstring for recoBox"""

    def __init__(self, owner, name, product, producers):
        super(recoBox, self).__init__()
        self._label = QtGui.QLabel()
        self._name = name
        self._label.setText(self._name.capitalize() + ": ")
        self._box = ComboBoxWithKeyConnect()
        self._box.activated[str].connect(self.emitSignal)
        self._product = product
        if producers == None:
            self._box.addItem("--None--")
        else:
            self._box.addItem("--Select--")
            for producer in producers:
                self._box.addItem(producer)

        self._box.connectOwnerKPE(owner.keyPressEvent)

        # This is the widget itself, so set it up
        self._layout = QtGui.QVBoxLayout()
        self._layout.addWidget(self._label)
        self._layout.addWidget(self._box)
        self.setLayout(self._layout)

    def keyPressEvent(self, e):
        self._box.keyPressEvent(e)
        super(recoBox, self).keyPressEvent(e)

    def emitSignal(self, text):
        self.activated.emit(text)

    def product(self):
        return self._product

    def name(self):
        return self._name





from .database import recoBase3D
from pyqtgraph.Qt import QtGui, QtCore
import numpy
import pyqtgraph.opengl as gl

import larcv

class cluster3d(recoBase3D):

    """docstring for cluster3d"""

    def __init__(self):
        super(cluster3d, self).__init__()
        self._product_name = 'cluster3d'

        self._listOfClusters = []

        self._gl_voxel_mesh = None
        self._id_summed_charge = []
        self._assigned_colors = []

        self._meta = None

        self._box_template = numpy.array([[ 0 , 0, 0],
                                          [ 1 , 0, 0],
                                          [ 1 , 1, 0],
                                          [ 0 , 1, 0],
                                          [ 0 , 0, 1],
                                          [ 1 , 0, 1],
                                          [ 1 , 1, 1],
                                          [ 0 , 1, 1]],
                                         dtype=float)


        self._faces_template = numpy.array([[0, 1, 2],
                                            [0, 2, 3],
                                            [0, 1, 4],
                                            [1, 5, 4],
                                            [1, 2, 5],
                                            [2, 5, 6],
                                            [2, 3, 6],
                                            [3, 6, 7],
                                            [0, 3, 7],
                                            [0, 4, 7],
                                            [4, 5, 7],
                                            [5, 6, 7]])

        # Defining the cluster3d colors:
        self._clusterColors = [
            (0, 147, 147, 125),    # dark teal
            (0, 0, 252, 125),      # bright blue
            (156, 0, 156, 125),    # purple
            (255, 0, 255, 125),    # pink
            (255, 0, 0, 125),      # red
            (175, 0, 0, 125),      # red/brown
            (252, 127, 0, 125),    # orange
            (102, 51, 0, 125),     # brown
            (100, 253, 0, 125),    # bright green
            (255, 248, 202, 125),  # tan
            (255, 248, 202, 125),  # tan
            (255, 148, 241, 125),  # pink
        ]


    # this is the function that actually draws the cluster.
    def drawObjects(self, view_manager, io_manager, meta):

        #Get the list of sparse3d sets:
        event_cluster3d = io_manager.get_data(self._product_name, str(self._producerName))

        cluster_set = event_cluster3d.at(0)
        self._meta = cluster_set.meta() 



        self._id_summed_charge = []
        self._assigned_colors = []

        _color_index = 0


        # # This section draws clusters onto the environment:
        for cluster in cluster_set.as_vector():

            _this_id_summed_charge = dict()
            for voxel in cluster.as_vector():
                if voxel.id() >= self._meta.total_voxels():
                    # print("continuing")
                    continue
                if voxel.id() in _this_id_summed_charge:
                    _this_id_summed_charge[voxel.id()] += voxel.value()
                else:
                    _this_id_summed_charge.update({voxel.id() : voxel.value()})

            self._id_summed_charge.append(_this_id_summed_charge)
            self._assigned_colors.append(self._clusterColors[_color_index])


            _color_index += 1
            if _color_index >= len(self._clusterColors):
                _color_index = 0


        # The last cluster is the 'leftover' depositions
        # Force it's color to white everytime:
        self._assigned_colors[-1] = (255, 255, 255, 125)

        self.redraw(view_manager)



    def redraw(self, view_manager):


        if self._gl_voxel_mesh is not None:
            view_manager.getView().removeItem(self._gl_voxel_mesh)
            self._gl_voxel_mesh = None

        verts, faces, colors = self.buildTriangleArray(self._id_summed_charge,
                                                       self._assigned_colors,
                                                       view_manager)


        #make a mesh item:
        mesh = gl.GLMeshItem(vertexes=verts,
                             faces=faces,
                             faceColors=colors,
                             smooth=False)
        # mesh.setGLOptions("additive")
        self._gl_voxel_mesh = mesh
        view_manager.getView().addItem(self._gl_voxel_mesh)

    def buildTriangleArray(self, id_summed_charge, assigned_colors, view_manager):


        n_voxels = 0
        for cluster in id_summed_charge:
            n_voxels += len(cluster)

        # # # Allocate enough memory for all 3 numpy arrays upfront:

        verts = numpy.zeros((n_voxels*8,3))
        faces = numpy.zeros((n_voxels*12,3), dtype=numpy.int)
        colors = numpy.zeros((n_voxels*12,4), dtype=numpy.int)


        i = 0
        for cluster, color in zip(id_summed_charge, assigned_colors):

            for voxel_id in cluster:

                # Don't draw this pixel if it's below the threshold:
                if cluster[voxel_id] < view_manager.getLevels()[0]:
                    continue

                colors[12*i: 12*(i+1)] = color
                faces[12*i:12*(i+1)] = self._faces_template + 8*i
                this_verts = self.makeBox(voxel_id, self._meta)
                verts[8*i:8*(i+1)] = this_verts

                i += 1

        colors = colors[0:i*12]
        faces  = faces[0:i*12]
        verts  = verts[0:i*8]


        return verts, faces, colors

    def makeBox(self, voxel_id, meta):
        verts_box = numpy.copy(self._box_template)
        #Scale all the points of the box to the right voxel size:
        verts_box[:,0] *= meta.voxel_dimensions(0)
        verts_box[:,1] *= meta.voxel_dimensions(1)
        verts_box[:,2] *= meta.voxel_dimensions(2)


        #Shift the points to put the center of the cube at (0,0,0)
        verts_box[:,0] -= 0.5*meta.voxel_dimensions(0)
        verts_box[:,1] -= 0.5*meta.voxel_dimensions(1)
        verts_box[:,2] -= 0.5*meta.voxel_dimensions(2)
        
        #Move the points to the right coordinate in this space
        verts_box[:,0] += meta.position(voxel_id, 0) - meta.origin(0)
        verts_box[:,1] += meta.position(voxel_id, 1) - meta.origin(1)
        verts_box[:,2] += meta.position(voxel_id, 2) - meta.origin(2)


        # color_arr = numpy.ndarray((12, 4))
        # color_arr[:] = [1,1,1,1]

        return verts_box



    def clearDrawnObjects(self, view_manager):
        if self._gl_voxel_mesh is not None:
            view_manager.getView().removeItem(self._gl_voxel_mesh)
            self._gl_voxel_mesh = None
        self._gl_voxel_mesh = None
        self._meta = None
        self._id_summed_charge = []
        self._assigned_colors = []

    def refresh(self, view_manager):
        self.redraw(view_manager)


from pyqtgraph.Qt import QtCore

from .event_meta import event_meta

import larcv

class evd_manager_base(QtCore.QObject):

    eventChanged = QtCore.pyqtSignal()
    drawFreshRequested = QtCore.pyqtSignal()
    metaRefreshed = QtCore.pyqtSignal(event_meta)


    def __init__(self, config, _file=None):
        super(evd_manager_base, self).__init__()
        QtCore.QObject.__init__(self)
        self._config = config
        self.init_manager(_file)

    def init_manager(self, _file):
        # For the larcv manager, using the IOManager to get at the data
        self._driver =  larcv.ProcessDriver("ProcessDriver")
        self._driver.configure(self._config)

        # Meta keeps track of information about number of planes, visible
        # regions, etc.:
        self._meta = event_meta()


        # Drawn classes is a list of things getting drawn, as well.
        self._drawnClasses = dict()

        self._keyTable = dict()


        if _file != None:
            # flist=larcv.VectorOfString()
            if type(_file) is list:
            #     for f in _file: flist.push_back(f)
                self._driver.override_input_file(_file)
            else:
                flist = []
                flist.append(_file)
                self._driver.override_input_file(flist)

        self._driver.initialize()

        self._driver.process_entry()
        # self._driver.set_id(run,subrun,event)

        self.refresh_meta()


    def refresh_meta(self):
        # # Read in any of the image2d products if none is specified.
        # # Use it's meta info to build up the meta for the viewer
        # _producers = self._driver.io().producer_list('image2d')
        # if 'wire' in _producers:
        #     _producer = 'wire'
        # else:
        #     _producer = _producers[-1]
        # _event_image2d = self._driver.io().get_data('image2d',_producer)
        

        # Meta information can come from either image2d, sparse2d or cluster2d
        # It's pull from image2d by default, from the first producer.
        # it goes to sparse2d next, then cluster2d

        product = "image2d"
        producers = self._driver.io().producer_list(product)



        if len(producers) == 0:
            product = "sparse2d"
            producers = self._driver.io().producer_list(product)
        if len(producers) == 0:
            product = "cluster2d"
            producers = self._driver.io().producer_list(product)
        
        if len(producers) == 0:
            raise Exception("No Meta avialable to define viewer boundaries")


        producer = producers[-1]


        meta_vec = []

        data = self._driver.io().get_data(product, producer)
        # if product == "image2d":
        #     data = larcv.EventImage2D.to_image2d(data)
        # if product == "sparse2d":
        #     data = larcv.EventSparseTensor2D.to_sparse_tensor(data)            
        # if product == "cluster2d":
        #     data = larcv.EventSparseCluster2D.to_sparse_cluster(data)

        for data_obj in data.as_vector():
            meta_vec.append(data_obj.meta())

        self._meta.refresh(meta_vec)

    def meta(self):
        return self._meta

    # This function will return all producers for the given product
    def getProducers(self, product):
        if self._driver.io() is not None:
            return self._driver.io().producer_list(product)

    # This function returns the list of products that can be drawn:
    def getDrawableProducts(self):
        return self._drawableItems.getDict()

    # override the run,event,subrun functions:
    def run(self):
        if self._driver.io() is None:
            return 0
        return self._driver.event_id().run()

    def event(self):
        if self._driver.io() is None:
            return 0
        return self._driver.event_id().event()

    def subrun(self):
        if self._driver.io() is None:
            return 0
        return self._driver.event_id().subrun()

    # def internalEvent(self):
    def entry(self):
        if self._driver.io() is not None:
            return self._driver.io().current_entry()
        else:
            return -1

    def n_entries(self):
        if self._driver.io() is not None:
            return self._driver.io().get_n_entries()
        else:
            return 0

    # override the functions from manager as needed here
    def next(self):
        if self.entry() + 1 < self.n_entries():
            # print self._driver.event()
            self._driver.clear_entry()            
            self._driver.process_entry(self.entry() + 1)
            self.eventChanged.emit()
        else:
            print("On the last event, can't go to next.")

    def prev(self):
        if self.entry != 0:
            self._driver.clear_entry()            
            self._driver.process_entry(self.entry() - 1)
            self.eventChanged.emit()
        else:
            print("On the first event, can't go to previous.")

    def go_to_entry(self, entry):
        if entry >= 0 and entry < self.n_entries():
            self._driver.clear_entry()            
            self._driver.process_entry(entry)
            self.eventChanged.emit()
        else:
            print("Can't go to entry {}, entry is out of range.".format(entry))

    def range(self, plane):
        # To get the range, we ask for the image meta and use it:
        return self._meta.range(plane)

    def n_views(self):
        return self._meta.n_views()



try:
    import pyqtgraph.opengl as gl
except:
    print("ERROR: Must have opengl for this display.")

from .gui3D import gui3D
from pyqtgraph.Qt import QtGui, QtCore
from manager import evd_manager_3D

from .recobox import recoBox

# Inherit the basic gui to extend it
# override the gui to give the lariat display special features:


class evdgui3D(gui3D):

    """special larlite gui for 3D"""

    def __init__(self):
        super(evdgui3D, self).__init__()


    def connect_manager(self, _manager):
        self._event_manager = _manager
        self._event_manager.eventChanged.connect(self.update)
        self._view_manager.refreshColors.connect(self.refreshColors)

    # override the initUI function to change things:
    def initUI(self):
        super(evdgui3D, self).initUI()
        # Change the name of the labels for lariat:
        self.update()

    def refreshColors(self):
        self._event_manager.refreshColors(self._view_manager)

    # This function sets up the eastern widget
    def getEastLayout(self):
        # This function just makes a dummy eastern layout to use.
        label1 = QtGui.QLabel("LArCV")
        label2 = QtGui.QLabel("EVD 3D")
        font = label1.font()
        font.setBold(True)
        label1.setFont(font)
        label2.setFont(font)

        self._eastWidget = QtGui.QWidget()
        # This is the total layout
        self._eastLayout = QtGui.QVBoxLayout()
        # add the information sections:
        self._eastLayout.addWidget(label1)
        self._eastLayout.addWidget(label2)
        self._eastLayout.addStretch(1)
        
        # self._paramsDrawBox = QtGui.QCheckBox("Draw Params.")
        # self._paramsDrawBox.stateChanged.connect(self.paramsDrawBoxWorker)
        # self._eastLayout.addWidget(self._paramsDrawBox)
        # self._eastLayout.addStretch(1)

        # Now we get the list of items that are drawable:
        drawableProducts = self._event_manager.getDrawableProducts()
        self._listOfRecoBoxes = []
        for product in drawableProducts:
            thisBox = recoBox(self,
                              product,
                              drawableProducts[product][1],
                              self._event_manager.getProducers(
                                  drawableProducts[product][1]))
            self._listOfRecoBoxes.append(thisBox)
            thisBox.activated[str].connect(self.recoBoxHandler)
            self._eastLayout.addWidget(thisBox)

        self._eastLayout.addStretch(2)

        self._eastWidget.setLayout(self._eastLayout)
        self._eastWidget.setMaximumWidth(150)
        self._eastWidget.setMinimumWidth(100)
        return self._eastWidget

    def drawableProductsChanged(self):
        # self.removeItem(self._eastLayout)
        self._eastWidget.close()
        east = self.getEastLayout()
        self.slave.addWidget(east)
        self.update()


    def recoBoxHandler(self, text):
        sender = self.sender()
        # print sender.product(), "was changed to" , text
        if text == "--Select--" or text == "--None--":
            self._event_manager.redrawProduct(sender.name(), None, self._view_manager)
            return
        else:
            self._event_manager.redrawProduct(sender.name(), text, self._view_manager)




try:
    import pyqtgraph.opengl as gl
except:
    print("Error, must have open gl to use this viewer.")
    exit(-1)

from pyqtgraph.Qt import QtGui, QtCore
import pyqtgraph as pg
import numpy as np
import math

class viewport3D(gl.GLViewWidget):

    quitRequested = QtCore.pyqtSignal()
    keyPressSignal = QtCore.pyqtSignal(QtGui.QKeyEvent)
    viewChanged = QtCore.pyqtSignal()


    def __init__(self):
        super(viewport3D, self).__init__()
        # add a view box, which is a widget that allows an image to be shown
        # add an image item which handles drawing (and refreshing) the image
        self.setBackgroundColor((0,0,0,255))

        self._background_items = []


        # self.pan(0,0,self._geometry.length())
        self.show()

    def updateMeta(self,meta):

        _len_x = meta.max_x() - meta.min_x()
        _len_y = meta.max_y() - meta.min_y()
        _len_z = meta.max_z() - meta.min_z()
    


        self.setCenter((0,0,0))
    
        for _item in self._background_items:
            self.removeItem(_item)
            self._background_items = []
    
        # This section prepares the 3D environment:
        # Add an axis orientation item:
        self._axis = gl.GLAxisItem()
        # self._axis.setSize(x=_len_x, y=_len_y, z=_len_z)
        self._axis.setSize(x=_len_x, y=0.25*_len_y, z=0.25*_len_z)
        self.addItem(self._axis)
    
        # Add a set of grids along x, y, z:
        self._xy_grid = gl.GLGridItem()
        # Set the grid size to the number of voxels in x and y:
        self._xy_grid.setSize(x=meta.n_voxels_x(), y=meta.n_voxels_y(), z=0.0)
        # Scale the grid to the correct size:
        self._xy_grid.scale(x=meta.size_voxel_x(), y=meta.size_voxel_y(), z=1.0)
        self._xy_grid.translate(_len_x*0.5, _len_y * 0.5, 0.0)

        self._yz_grid = gl.GLGridItem()
        self._yz_grid.setSize(x=meta.n_voxels_z(), y=meta.n_voxels_y())
        self._yz_grid.scale(x=meta.size_voxel_z(), y=meta.size_voxel_y(), z=1.0)
        self._yz_grid.rotate(-90, 0, 1, 0)
        self._yz_grid.translate(0, _len_y*0.5, _len_z*0.5)

        self._xz_grid = gl.GLGridItem()
        self._xz_grid.setSize(x=meta.n_voxels_x(), y=meta.n_voxels_z())
        self._xz_grid.scale(x=meta.size_voxel_x(), y=meta.size_voxel_z(), z=1.0)
        self._xz_grid.rotate(90, 1, 0, 0)
        self._xz_grid.translate(_len_x*0.5, 0, _len_z*0.5)
    
        self.addItem(self._xy_grid)
        self.addItem(self._yz_grid)
        self.addItem(self._xz_grid)
    
        self._background_items.append(self._axis)
        self._background_items.append(self._xy_grid)
        self._background_items.append(self._yz_grid)
        self._background_items.append(self._xz_grid)
    
        # Move the center of the camera to the center of the view:
        self.pan(_len_x*0.5, _len_y * 0.5, _len_z*0.5)

    def setCenter(self, center):
        if len(center) != 3:
            return
        cVec = QtGui.QVector3D(center[0],center[1],center[2])
        self.opts['center'] = cVec
        self.update()




    def worldCenter(self):
        return self.opts['center']

  # def drawBox(self, xmin, ymin, zmin, xmax, ymax, zmax):

  #   # Draw lines from each corner to the
  #   # 3 corners it connects with
  #   self.drawLine((xmin,ymin,zmin),(xmin,ymax,zmin))
  #   self.drawLine((xmin,ymin,zmin),(xmax,ymin,zmin))
  #   self.drawLine((xmax,ymin,zmin),(xmax,ymax,zmin))
  #   self.drawLine((xmin,ymax,zmin),(xmax,ymax,zmin))

  #   self.drawLine((xmin,ymin,zmax),(xmin,ymax,zmax))
  #   self.drawLine((xmin,ymin,zmax),(xmax,ymin,zmax))
  #   self.drawLine((xmax,ymin,zmax),(xmax,ymax,zmax))
  #   self.drawLine((xmin,ymax,zmax),(xmax,ymax,zmax))

  #   self.drawLine((xmin,ymin,zmin),(xmin,ymin,zmax))
  #   self.drawLine((xmax,ymax,zmin),(xmax,ymax,zmax))
  #   self.drawLine((xmin,ymax,zmin),(xmin,ymax,zmax))
  #   self.drawLine((xmax,ymin,zmin),(xmax,ymin,zmax))


    def getAzimuth(self):
        return self.opts['azimuth']


    def getElevation(self):
        return self.opts['elevation']

    def setCameraPos(self,pos):
        # calling set camera with pos doesn't actually do anything.  Convert spherical coordinates:
        if pos is not None:
            # Convert to relative coordinates to always leave the world center as the center point
            worldCenter = self.opts['center']
            # Check the type:
            if type(worldCenter) is QtGui.QVector3D:
                X = pos[0] - worldCenter.x()
                Y = pos[1] - worldCenter.y()
                Z = pos[2] - worldCenter.z()
            else:
                X = pos[0] - worldCenter[0]
                Y = pos[1] - worldCenter[1]
                Z = pos[2] - worldCenter[2]

            distance = X**2 + Y**2 + Z**2
            distance = math.sqrt(distance)
            if X != 0:
                azimuth = math.atan2(Y,X)
            else:
                azimuth = math.pi
                if Y < 0:
                    azimuth = -1 * azimuth
            if distance != 0:
                elevation = math.asin(Z / distance)
            else:
                elevation = math.copysign(Z)
            azimuth *= 180./math.pi
            elevation *= 180./math.pi
            self.setCameraPosition(distance=distance,elevation=elevation,azimuth=azimuth)
            self.viewChanged.emit()

    def keyPressEvent(self,e):
        if e.key() == QtCore.Qt.Key_C:
            # print "C was pressed"
            if e.modifiers() and QtCore.Qt.ControlModifier :
                self.quitRequested.emit()
                return
        elif e.modifiers():
            if QtCore.Qt.ShiftModifier :
                if e.key() == QtCore.Qt.Key_Up:
                    # This is supposed to pan upwards in the view
                    self.pan(0,20,0,True)
                if e.key() == QtCore.Qt.Key_Down:
                    self.pan(0,-20,0,True)
                if e.key() == QtCore.Qt.Key_Left:
                    self.pan(20,0,0,True)
                if e.key() == QtCore.Qt.Key_Right:
                    self.pan(-20,0,0,True)
        else:
            super(viewport3D,self).keyPressEvent(e)
       

        # Pass this signal to the main gui, too
        self.keyPressSignal.emit(e)

    def orbit(self, azim, elev):
        super(viewport3D, self).orbit(azim, elev)
        self.viewChanged.emit()

    def pan(self, dx, dy, dz, relative=False):
        super(viewport3D, self).pan(dx, dy, dz, relative)
        self.viewChanged.emit()

    def wheelEvent(self, ev):
        super(viewport3D, self).wheelEvent(ev)
        self.viewChanged.emit()

  

  # def drawLine(self,point1,point2):

  #   x = np.linspace(point1[0],point2[0],100)
  #   y = np.linspace(point1[1],point2[1],100)
  #   z = np.linspace(point1[2],point2[2],100)

  #   pts = np.vstack([x,y,z]).transpose()
  #   line = gl.GLLinePlotItem(pos=pts)
  #   self.addItem(line)

  # def mouseMoveEvent(self, pos):
    # print pos




from .database import recoBase
from pyqtgraph.Qt import QtGui, QtCore
import pyqtgraph as pg

class bbox2d(recoBase):

    """docstring for cluster"""

    def __init__(self):
        super(bbox2d, self).__init__()
        self._product_name = 'bbox2d'

    # this is the function that actually draws the cluster.
    def drawObjects(self, view_manager, io_manager, meta):


        event_bbox2d = io_manager.get_data(self._product_name, str(self._producerName))


        self._drawnObjects = []
        for plane, view in view_manager.getViewPorts().items():
            # get the plane
            # thisPlane = view.plane()
            self._drawnObjects.append([])

            collection = event_bbox2d.at(plane)



            for bbox2d in collection.as_vector():
                
                print(bbox2d)

                # A QRect can be constructed with a set of left, top, width and height integers, 
                # or from a QPoint and a QSize.

                c  = bbox2d.centroid()
                hl = bbox2d.half_length()

                r = QtGui.QGraphicsRectItem(c[0] - hl[0], c[1] - hl[1], 2*hl[0], 2*hl[1])


                # particle = event_bbox2d.at(i)
                # bounding_box = particle.boundingbox_2d(plane)

                # left = meta.wire_to_col(bounding_box.min_y(), plane)
                # right = meta.wire_to_col(bounding_box.max_y(), plane)
                # top = meta.time_to_row(bounding_box.min_x(), plane)
                # bottom = meta.time_to_row(bounding_box.max_x(), plane)

                # #r = QtGui.QGraphicsRectItem(bottom, left, (top - bottom), (right-left))
                # r = QtGui.QGraphicsRectItem(left, bottom, (right-left), (top - bottom))
                r.setPen(pg.mkPen('r'))
                r.setBrush(pg.mkColor((0,0,0,0)))
                self._drawnObjects[plane].append(r)
                view._plot.addItem(r)

        return

    # def clearDrawnObjects(self, view_manager):
    #     i_plane = 0
    #     # erase the clusters
    #     for plane in self._listOfClusters:
    #         view = view_manager.getViewPorts()[i_plane]
    #         i_plane += 1
    #         for cluster in plane:
    #             cluster.clearHits(view)


    #     self._listOfClusters = []



from pyqtgraph.Qt import QtCore
import datatypes

from .evd_manager_base import evd_manager_base

class evd_manager_2D(evd_manager_base):

    '''
    Class to handle the 2D specific aspects of viewer
    '''

    def __init__(self, config, _file=None):
        super(evd_manager_2D, self).__init__(config, _file)
        self._drawableItems = datatypes.drawableItems()

    # this function is meant for the first request to draw an object or
    # when the producer changes
    def redrawProduct(self, product, producer, view_manager):
        
        # print "Received request to redraw ", product, " by ",producer
        # First, determine if there is a drawing process for this product:  
        
        if producer is None:
            if product in self._drawnClasses:
                self._drawnClasses[product].clearDrawnObjects(view_manager)
                self._drawnClasses.pop(product)
            return
        if product in self._drawnClasses:
            self._drawnClasses[product].setProducer(producer)
            self._drawnClasses[product].clearDrawnObjects(view_manager)
            self._drawnClasses[product].drawObjects(view_manager, self._driver.io(), self.meta())
            return

        # Now, draw the new product
        if product in self._drawableItems.getListOfTitles():
            # drawable items contains a reference to the class, so
            # instantiate it
            drawingClass=self._drawableItems.getDict()[product][0]()

            drawingClass.setProducer(producer)
            self._drawnClasses.update({product: drawingClass})

            # Need to process the event
            drawingClass.drawObjects(view_manager, self._driver.io(), self.meta())


    def clearAll(self, view_manager):
        for recoProduct in self._drawnClasses:
            self._drawnClasses[recoProduct].clearDrawnObjects(view_manager)
        # self.clearTruth()

    def drawFresh(self, view_manager):

        self.clearAll(view_manager)
        # Draw objects in a specific order defined by drawableItems
        order = self._drawableItems.getListOfTitles()
        # self.drawTruth()
        for item in order:
            if item in self._drawnClasses:
                self._drawnClasses[item].drawObjects(view_manager, self._driver.io(), self.meta())



    def getPlane(self, plane):
        if self._drawWires:
            return self._wireDrawer.getPlane(plane)



from .database import recoBase
from pyqtgraph.Qt import QtGui, QtCore
import pyqtgraph, numpy
from .connectedObjects import connectedBox, boxCollection

import larcv

class sparse2d(recoBase):

    """docstring for sparse2d"""

    def __init__(self):
        super(sparse2d, self).__init__()
        self._product_name = 'sparse2d'

        self._listOfClusters = []

        # Defining the sparse2d colors:
        self._clusterColors = [
            (0, 147, 147, 125),  # dark teal
            (0, 0, 252, 125),   # bright blue
            (156, 0, 156, 125),  # purple
            (255, 0, 255, 125),  # pink
            (255, 0, 0, 125),  # red
            (175, 0, 0, 125),  # red/brown
            (252, 127, 0, 125),  # orange
            (102, 51, 0, 125),  # brown
            (127, 127, 127, 125),  # dark gray
            (210, 210, 210, 125),  # gray
            (100, 253, 0, 125)  # bright green
        ]

    # this is the function that actually draws the cluster.
    def drawObjects(self, view_manager, io_manager, meta):

        #Get the list of sparse2d sets:
        sparse_2d_set = io_manager.get_data(self._product_name, str(self._producerName))
        # sparse_2d_set = larcv.EventSparseTensor2D.to_sparse_tensor(sparse_2d_set)
        # if self._producerName in io_manager.producer_list(self._product_name):
        #     hasROI = True
        # else:
        #     hasROI = False

        # if hasROI:
        #     event_roi = io_manager.get_data(self._product_name, str(self._producerName))

        for plane, view in view_manager.getViewPorts().items():
            self._drawnObjects.append([])

            colorIndex = 0

            active_map = view._activeMap
            cols = []
            vals = []
            for val, rgba in active_map['ticks']:
                cols.append(numpy.asarray(rgba) / 255.)
                vals.append(val)

            cmap = pyqtgraph.ColorMap(vals, cols)

            # n = len(uids); 
            # colors = cmap(range(n), bytes = True);


            # Get the sparse2d clusters for this plane:
            # try:
            voxelset = sparse_2d_set.sparse_tensor(plane)

            # Voxelset doesn't manager memory for just indexes or just values
            # Therefore, we keep a reference to these.
            # If we don't, the memory underneath for the std::vectors will
            # be freed and anything might end up there.  This will corrupt the 
            # numpy arrays below since they share the memory.
            meta = voxelset.meta()

            # cpp_indexes = voxelset.indexes()
            # cpp_values = voxelset.values()

            # indexes = larcv.as_ndarray_sizet(cpp_indexes)
            # values  = larcv.as_ndarray_float(cpp_values)

            indexes = voxelset.indexes()
            values = voxelset.values()

            # Reject all out-of-bounds indexes:
            in_bounds = indexes < meta.total_voxels()
            oob = indexes >= meta.total_voxels()
            indexes = indexes[in_bounds]
            values  = values[in_bounds]

            dims = [meta.number_of_voxels(0), meta.number_of_voxels(1) ]
            x, y = numpy.unravel_index(indexes, dims)

            y = y.astype(float) + 0.5
            x = x.astype(float) + 0.5



            this_item = pyqtgraph.ScatterPlotItem()

            this_item.setData(x, y, size=1, pxMode=False, symbol='s', pen=None)

            view._plot.addItem(this_item)
            self._drawnObjects[plane].append(this_item)

